import re
import ast
from abc import ABC, ABCMeta, abstractmethod
import os.path
import sys
import importlib
import json

from .task import Task
from .task import serialize_kwargs
from .messages import StoreMsg
from .serialization import error_to_string, string_to_error

# ==== Persisting ====
#
# Loading and Saving results
#
# -- registering persisters --
#
# Each persister is identified by a path. To figure out which persister should be used to save/load a result,
# each task has a path to a persister. The Persister_factory takes the path and gives the appropriate
# persister back. Persisters can also be defined by endusers and are automatically registered.


_persisters = {}


def register_persister(persister, path):
    _persisters[path] = persister


def persister_factory(path):
    """Return the persister registered via path"""
    m = re.match("(?P<name>.+)\\((?P<args>.*)\\)", path)
    if m.group("args"):
        args = ast.literal_eval('(' + m.group('args') + ')')
    else:
        args = []
    return _persisters[m.group("name")](*args)


def PersisterMapper(f):
    persister_path = getattr(f, "_persister", "remoter.InbasePersister()")
    return persister_factory(persister_path)


# -- default persisters --


class Persister(ABC):
    """
    Persists task results and generated metadata in form of a `StoreMsg`

    After calculating the result of a task, a Persister persists the result of the task.
    The metadata how the result is persisted and how it can be obtained again is saved
    in a `StoreMsg`. The StoreMsg is passed on to a store, which distributes it to all
    other stores and saves it.
    """

    path = "abstract"

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        register_persister(cls, cls.path)

    def to_msg(self, task: "Task") -> StoreMsg:
        """Persist task result and conver to StoreMsg"""
        pass

    def from_msg(self, msg: StoreMsg) -> "Task":
        """Load a task result from a storemsg"""
        pass


class InbasePersister(Persister):
    """Persists results as a string in the sqlite task database"""

    path = "remoter.InbasePersister"

    def __init__(self):
        self.path = self.path + "()"

    def to_msg(self, task):
        result = task.result
        msg = StoreMsg(
            task.datetime,
            task.function_module,
            task.function_name,
            task.function_hash,
            task.kwargs,
            result,
            task.task_hash,
            self.path,
            task.failed,
            task.time
        )
        return msg

    def from_msg(self, msg):
        result = msg.result
        return Task(
            self,
            None,
            msg.datetime,
            msg.function_module,
            msg.function_name,
            msg.function_hash,
            msg.kwargs,
            msg.result,
            msg.task_hash,
            True,
            time=msg.time
        )


class FunctionPersister(Persister):
    """Persists results with user defined functions"""

    path = "remoter.FunctionPersister"

    def __init__(self, load_function, save_function, load_kwargs, save_kwargs):
        self.path = self.path + f"('{load_function}', '{save_function}', '{load_kwargs}', '{save_kwargs}')"
        self._load_function = load_function
        self._save_function = save_function
        self._load_kwargs = json.loads(load_kwargs)
        self._save_kwargs = json.loads(save_kwargs)
        self._modules = {}

    def _get_function(self, name):
        module, functionname = name.rsplit(".", 1)
        if module not in self._modules:
            self._modules[module] = importlib.import_module(module)
        f = getattr(self._modules[module], functionname)
        return f

    def _save(self, result):
        return self._get_function(self._save_function)(result, **self._save_kwargs)

    def _load(self, result):
        return self._get_function(self._load_function)(result, **self._load_kwargs)

    def to_msg(self, task):
        result = self._save(task.result)
        msg = StoreMsg(
            task.datetime,
            task.function_module,
            task.function_name,
            task.function_hash,
            task.kwargs,
            result,
            task.task_hash,
            self.path,
            task.failed,
            task.time
        )
        return msg

    def from_msg(self, msg):
        result = self._load(msg.result)
        return Task(
            self,
            None,
            msg.datetime,
            msg.function_module,
            msg.function_name,
            msg.function_hash,
            msg.kwargs,
            result,
            msg.task_hash,
            True,
            msg.time
        )


def save(fn, **kwargs):
    """Indicate which function to use to save the result of this function

    Used by the FunctionPersister
    """
    module = fn.__module__ if fn.__module__ != "__main__" else os.path.split(sys.argv[0])[-1].replace(".py", "")
    n = module + "." + fn.__name__

    def wrapper(f):
        f.__remoter_save = n
        f.__remoter_save_kwargs = json.dumps(kwargs)
        if hasattr(f, "__remoter_load") and hasattr(f, "__remoter_save"):
            l, s = f.__remoter_load, f.__remoter_save
            lkwargs, skwargs = f.__remoter_load_kwargs, f.__remoter_save_kwargs
            f._persister = f"remoter.FunctionPersister('{l}', '{s}', '{lkwargs}', '{skwargs}')"
        return f

    return wrapper


def load(fn, **kwargs):
    """Indicate which function to use to load the result of this function

    Used by the FunctionPersister
    """
    module = fn.__module__ if fn.__module__ != "__main__" else os.path.split(sys.argv[0])[-1].replace(".py", "")
    n = module + "." + fn.__name__

    def wrapper(f):
        f.__remoter_load = n
        f.__remoter_load_kwargs = json.dumps(kwargs)
        if hasattr(f, "__remoter_load") and hasattr(f, "__remoter_save"):
            l, s = f.__remoter_load, f.__remoter_save
            lkwargs, skwargs = f.__remoter_load_kwargs, f.__remoter_save_kwargs
            f._persister = f"remoter.FunctionPersister('{l}', '{s}', '{lkwargs}', '{skwargs}')"
        return f

    return wrapper


class ErrorPersister(Persister):
    """Persists an error as a string in the sqlite task database"""

    path = "remoter.ErrorPersister"

    def __init__(self):
        self.path = self.path + "()"

    def to_msg(self, task):
        msg = StoreMsg(
            task.datetime,
            task.function_module,
            task.function_name,
            task.function_hash,
            task.kwargs,
            error_to_string(task.exception),
            task.task_hash,
            self.path,
            task.failed,
            task.time
        )
        return msg

    def from_msg(self, msg):
        result = msg.result
        return Task(
            self,
            None,
            msg.datetime,
            msg.function_module,
            msg.function_name,
            msg.function_hash,
            msg.kwargs,
            None,
            msg.task_hash,
            True,
            failed=True,
            exception=string_to_error(msg.result),
            time=msg.time
        )




import regex
import json
import phantombuster.plumbing as plumbing
import phantombuster.config_files

from .plumbing import pairwise
from phantombuster import bamindexer
from phantombuster import statistics
from phantombuster.io_ import read_file, write_parquet
from phantombuster.plumbing import collect

import os
import logging
from collections import defaultdict, namedtuple, Counter
import gzip

import numpy as np
import polars as pl


def index_group(group, every):
    if len(group.files) == 1:
        return index_file(group.files[0].path, every)
    else:
        idxs = [[0]* len(group.files)] + [[np.inf]* len(group.files)]
        return list(zip(*idxs))


def index_file(file, every):
    if file.endswith('.bam'):
        return index_bamfile(file, every)
    elif file.endswith('.fastq.gz'):
        # Indexing calculates the index of every `every`-th object.
        # For BAM files, the objects are BGZF blocks with a maximal size of 64KB.
        # For index_fastqgz_file, the objects are reads, which leads to a missmatch
        # between the indexing behavior for BAM files and fastq files.
        # As a workaround every gets multiplied here by 64000 to closer match
        # how indexing works for BAM files.
        return index_fastqgz_file(file, every*64000)

def index_bamfile(bamfile, every):
    with open(bamfile, mode="rb") as f:
        return bamindexer.index(f, every)


def _read_complete_read(f, read_size=1024):
    buffer = []
    lines = 0
    tries = 0
    while lines < 7 and tries < 100:
        read = f.read(read_size)
        if len(read) == 0:
            raise EOFError()
        lines += read.count(b'\n')
        buffer.append(read)
        tries += 1
    if tries >= 100:
        raise ValueError("Can't find reads in fastq file")
    complete_read = b''.join(buffer)
    lines = complete_read.split(b'\n')
    for line_idx in range(5):
        if len(lines[line_idx]) > 0 and lines[line_idx][0] == b'@'[0]:
            break
        else:
            pass
            #print('no @ in', lines[line_idx][0])
    else:
        raise EOFError()

    single_read = b'\n'.join(lines[line_idx:line_idx+4])
    offset = complete_read.find(single_read)
    return single_read, offset


def index_fastqgz_file(file, every):
    with gzip.open(file) as f:
        read, offset = _read_complete_read(f)
        total_readsize = len(read)
        reads_analysed = 1

        readsize = int(total_readsize / reads_analysed)

        idx = 0
        indices = []
        while True:
            try:
                f.seek(idx)
                read, offset = _read_complete_read(f, readsize)

                indices.append(idx+offset)
                idx = idx + offset + every

                total_readsize += len(read)
                reads_analysed += 1
                readsize = int(total_readsize / reads_analysed)

            except EOFError:
                break
        indices.append(np.inf)
    return indices


def demultiplex(input_files_file, outpath, outpath_stats, regex_file, barcode_hierarchy_file, debug, show_qc, scheduler):

    logger = logging.getLogger("phantombuster")

    input_groups = phantombuster.config_files.read_input_files_file(input_files_file)

    logger.info(f"Deduplicating {len(input_groups)} many groups")
    logger.info("Reading mappings")

    # read mapping files
    barcode_hierarchy = phantombuster.config_files.read_barcode_hierarchy_file(barcode_hierarchy_file)

    regex_dictionary = phantombuster.config_files.read_regex_file(regex_file)

    logger.info("Scheduling indexing %s group(s)", len(input_groups))
    warned_for_missing_workers = False
    if scheduler.query_workers() == 0:
        warned_for_missing_workers = True
        logger.warning("Work is getting scheduled and no workers are currently connected. Launch workers now.")

    idxs_promises = []
    for group in input_groups:
        # BGZF blocks per task, each block is <64KB so 20000 should roughly be 1.28 GB
        promise = scheduler.schedule(index_group, group, 20000)
        idxs_promises.append(promise)

    for promise in idxs_promises:
        r = 0
        waiting_time = 0
        waiting_notice_time = 10
        while isinstance(r, int):
            waiting_time += r
            if waiting_time > waiting_notice_time:
                waiting_time = 0
                number_of_workers = scheduler.query_workers()
                if number_of_workers == 0:
                    warned_for_missing_workers = True
                    logger.warning("Currently no workes are connected to the server. Server will make no progress without workers.")
                    if waiting_notice_time == 10:
                        waiting_notice_time = 60
                    elif waiting_notice_time == 60:
                        waiting_notice_time = 10*60
                else:
                    if warned_for_missing_workers:
                        logger.info(f"Currently {number_of_workers} many workers are connected.")
                        warned_for_missing_workers = False
                    waiting_notice_time = 10*60
            r = scheduler.wait(promise, return_after_wait=True)

    idxs = [scheduler.wait(promise, unpack=True) for promise in idxs_promises]

    logger.info("Parsing file(s) piecewise, %s many subtasks", sum(len(idx) for idx in idxs))
    logger.debug("Index is %s", idxs)
    outs = []
    for group, index in zip(input_groups, idxs):
        for section in pairwise(index):
            logger.debug("Scheduling deduplicate section %s", section)
            t = scheduler.schedule(plumbing.deduplicate_section, group, section, regex_dictionary, barcode_hierarchy, show_qc)
            outs.append(t)
    scheduler.wait_all(outs, unpack=False)

    logger.info("Extracted reads from individual files, combining %s many files", len(outs))
    new_outs = []
    while len(outs) > 1:
        while len(outs) > 1:
            a = outs.pop(0)
            b = outs.pop(0)
            logger.debug("Waiting on two files")
            scheduler.wait(a, unpack=False)
            scheduler.wait(b, unpack=False)
            logger.debug("Scheduling merging of two files, %s remaining.", len(outs))
            result = scheduler.schedule(plumbing.combine, [a, b], barcode_hierarchy)
            new_outs.append(result)
        if len(outs) == 1:
            logger.debug("Cannot merge, only one file remaining")
            new_outs.append(outs[0])
        logger.debug("One round done, there were %s many files, now there are %s", len(outs), len(new_outs))
        outs = new_outs
        new_outs = []

    logger.debug("Waiting for last combination round")
    out = scheduler.wait(outs[-1], unpack=True)
    if out is None or out[0] is None:
        logger.error("Not a single read was extracted. This might indicates a problem with the references or a missconfiguration of the regexes. Aborting. ")
        logger.debug('Result of combining all files is: %s', out)
        raise Exception("Not a single read extracted")

    write_parquet(out[0], outpath)

    with open(outpath_stats, mode="w") as f:
        json.dump(out[1], f)

    logger.info("Demultiplexing done")
    return out

def error_correct(samplefile, outfilename, threshold, barcode_hierarchy, project, remove_ambigious):

    logger = logging.getLogger('phantombuster')

    logger.debug(f"Reading table {samplefile}")
    df = pl.scan_parquet(samplefile)
    number_of_reads, number_of_lineages = collect(df.select(pl.col('reads').sum(), pl.len()))
    number_of_reads, number_of_lineages = number_of_reads[0], number_of_lineages[0]
    logger.debug(f"Table has {number_of_lineages} many rows and {number_of_reads} many reads")

    bc_idx = 0
    anchors = []

    while bc_idx < len(barcode_hierarchy) and barcode_hierarchy[bc_idx]["type"] == "reference":
        anchors.append(barcode_hierarchy[bc_idx]["name"])
        bc_idx += 1

    if len(anchors) > 0:
        logger.debug(f"Partitioning on reference type barcodes {anchors}")
        partitions = list(collect(df.group_by(anchors).agg()).iter_rows())
        logger.debug(f"Partitioning Done, created {len(partitions)} many partitions.")
    else:
        partitions = [None]

    scheduler = project.get_scheduler()
    with scheduler as scheduler:
        logger.debug('Distributing each partition to one worker')
        tasks = []
        for value in partitions:
            task = scheduler.schedule(error_correct_partition, samplefile, value, None, threshold, barcode_hierarchy, use_column=anchors, remove_ambigious=remove_ambigious)
            tasks.append(task)
        logger.debug('Distributed all partitions, waiting for workers')
        results = scheduler.gather(*tasks)
        logger.debug('Workers done, performing postprocessing')

    tables = [table for table, stats in results]
    tables = [table for table in tables if not collect(table.head(1)).is_empty()]
    tables = sorted(tables, key=lambda table: list(collect(table.head(1).select(anchors)).iter_rows())[0])

    corrected_table = pl.concat(tables)

    number_of_lineages_corrected = collect(df.select(pl.len()))["len"][0]
    number_of_reads_corrected = collect(corrected_table.select(pl.col("reads").sum()))["reads"][0]

    write_parquet(corrected_table, outfilename)

    return {"sample": samplefile,
            "lids_uncorrected": number_of_lineages,
            "lids_corrected": number_of_lineages_corrected,
            "reads_uncorrected": number_of_reads,
            "reads_corrected": number_of_reads_corrected}

def error_correct_partition(samplefile, lower_or_value, length_opt, threshold, barcode_hierarchy, use_column=False, remove_ambigious=True):

    logger = logging.getLogger("phantombuster")

    if lower_or_value is None:
        logger.info(f'Read table from {samplefile} completly')
        table = collect(pl.scan_parquet(samplefile))
    elif use_column:
        assert isinstance(use_column, list)
        assert isinstance(lower_or_value, list)
        value = lower_or_value
        table = pl.scan_parquet(samplefile).filter(*[pl.col(c) == v for c,v in zip(use_column, value)]).collect()
        logger.info(f'Read table from {samplefile} with {" ".join([f"{c} = {v}"  for c,v in zip(use_column, value)]) }')
    else:
        assert isinstance(lower_or_value, int)
        lower = lower_or_value
        assert length_opt is not None
        length = length_opt

        table = read_parquet(samplefile)
        table = table.slice(lower, length)
        logger.info(f'Read table from {samplefile} from {lower} with length {length}.')

    number_of_lineages = len(table)
    number_of_reads = table["reads"].sum()

    logger.info('Starting error correction')
    corrected_table = plumbing.error_correct(table, barcode_hierarchy, threshold)
    logger.info('Done error correction')

    if remove_ambigious:
        logger.info('Removing barcode sequences with ambigious characters (NRYSWKMBDHV).')
        corrected_table = plumbing.remove_ambigious(corrected_table, barcode_hierarchy)

    number_of_lineages_corrected = len(corrected_table)
    number_of_reads_corrected = corrected_table["reads"].sum()

    stats = {"lids_uncorrected": number_of_lineages,
            "lids_corrected": number_of_lineages_corrected,
            "reads_uncorrected": number_of_reads,
            "reads_corrected": number_of_reads_corrected}

    return corrected_table, stats

def _read_mappingfile(mappingfile):
    table = read_file(mappingfile)
    map = dict(zip(table["sample"].to_list(), table["group"].to_list()))
    return map

def hopping_removal(input_file, alpha_threshold, hopping_barcodes):
    logger = logging.getLogger("phantombuster")
    master_table = collect(read_file(input_file))

    number_of_lineages = len(master_table)
    number_of_reads = master_table["reads"].sum()

    logger.info("Starting Hopping Removal")
    threshold_list = []
    for barcode in hopping_barcodes:
        logger.info(f"Considering Barcode '{barcode}'")
        threshold = plumbing.calculate_hopping_threshold(master_table, barcode, alpha_threshold=alpha_threshold)
        threshold_list.append(threshold)
    master_table = plumbing.call_hopping(master_table, threshold_list)

    logger.info("Hopping Removal Done")

    number_of_lineages_corrected = len(master_table)
    number_of_reads_corrected = master_table["reads"].sum()

    stats = {"lids_before": number_of_lineages,
             "lids_after": number_of_lineages_corrected,
             "reads_before": number_of_reads,
             "reads_after": number_of_reads_corrected}

    return master_table, stats


def threshold(project, threshold_file):
    thresholds = pl.DataFrame(read_file(threshold_file))
    table = pl.read_parquet(project.hopping_removal_output_path)

    lids_before = len(table)
    reads_before = table.select(pl.col('reads').sum())['reads'][0]

    if len(thresholds.columns) == 1:
        table = table.filter(pl.col('reads') >= thresholds['threshold'][0])
    else:
        columns = table.columns
        table = table.join(thresholds, on=[col for col in table.columns if col != 'threshold' and col in thresholds.columns]).filter(pl.col('reads') >= pl.col('threshold'))
        table = table.select(columns)

    lids_after = len(table)
    reads_after = table.select(pl.col('reads').sum())['reads'][0]

    stats = {"lids_before": lids_before,
             "lids_after": lids_after,
             "reads_before": reads_before,
             "reads_after": reads_after}
    return table, stats





import sys
import hashlib
import subprocess
import os
import tempfile
import pytest
import shutil
import numpy.random
import numpy as np
import glob
from click.testing import CliRunner
import gzip
import regex
import polars as pl


from os.path import join as pjoin
import uuid
from collections import Counter
import datetime

from phantombuster import plumbing, analysis, core
import phantombuster.cli
import phantombuster.io_
import phantombuster.remoter
import phantombuster.plumbing
import phantombuster.statistics
from phantombuster.plumbing import table_to_counter
from phantombuster.plumbing import deduplicator_to_table
from phantombuster.plumbing import get_sectioned_iterator, pairwise
from phantombuster.porcelain import index_bamfile, index_fastqgz_file
from phantombuster.error_corrector import ErrorCorrector, Umi, UmiCount
from phantombuster.project import Project
from phantombuster.io_ import read_file
import phantombuster.config_files
import phantombuster.remoter.serialization
import regex

import zmq
from phantombuster.handler import PushHandler
import logging
import threading
import time



phantombuster.remoter.Scheduler.DEFAULT_WORKERS = 2

ALPHABET = list("ACGT")

phantombuster.cli.configure_logging(None, False)


def generate_barcode(length, k=1, alphabet=ALPHABET):
    barcodes = numpy.random.choice(ALPHABET, (k, length))
    return ["".join(barcode) for barcode in barcodes[:, ]]


def gen_bc(barcodes, weights=None, blocksize=1024):
    while True:
        block = np.random.choice(barcodes, blocksize, p=weights)
        for lid in block:
            yield lid


class ReadGenerator:

    def __init__(self, samples=10, genes=20):
        self.fixed = "GGGC"

        self.b2regex = "(?P<umi>[ACGT]{9})"
        self.bcregex = "(?P<grna>[ACGT]{11})"
        self.queryregex = "GGGC(?P<sample>[ACGT]{5})(?P<lid>[ACGT]{30})"

        self.samples = generate_barcode(5, samples)
        self.sample_dict = {s: f"Sample{i}" for i, s in enumerate(self.samples)}

        self.genes = generate_barcode(11, genes)
        self.gene_dict = {s: f"Gene{i}" for i, s in enumerate(self.genes)}

    def generate_reads(self, N, lids=12000, umis=1000):

        lids = generate_barcode(30, lids)
        umis = generate_barcode(9, umis)

        lids_pregen = gen_bc(lids)
        umis_pregen = gen_bc(umis)
        genes_pregen = gen_bc(self.genes)
        sample_pregen = gen_bc(self.samples)

        reads = []

        for i in range(N):
            lid = next(lids_pregen)
            umi = next(umis_pregen)
            gene = next(genes_pregen)
            sample = next(sample_pregen)
            read = make_read(f"{self.fixed}{sample}{lid}", umi, gene)
            reads.append(read)

        return reads


def make_read(query, b2, bc):
    query_qc = "F"*len(query)
    b2_qc = "F"*len(b2)
    bc_qc = "F"*len(bc)

    read = f"D00689:314:CBYRBANXX:1:1104:4980:1967\t4\t*\t0\t0\t*\t*\t0\t0\t{query}\t{query_qc}\tB2:Z:{b2}\tQ2:Z:{b2_qc}\tBC:Z:{bc}\tQT:Z:{bc_qc}\tRG:Z:CBYRBANXX.1\n".encode("utf-8")
    return read


def copy_file_efficiently(src, dst):
    with open(src, mode="br") as f_src:
        count = os.fstat(f_src.fileno()).st_size
        with open(dst, mode="bw") as f_dst:
            shutil.copyfile(src, dst)


# -- Unit Tests -- #

def test_unit_serialization_extensions_1():
    input_file = phantombuster.config_files.InputFile('a', 'b', 'c')
    serialized = phantombuster.remoter.serialization.serialize_value(input_file)
    deserialized = phantombuster.remoter.serialization.deserialize_value(serialized, None)

    assert deserialized == input_file


def test_unit_serialization_extensions_2():
    input_file = phantombuster.config_files.RegexDictionary({('', '', 'b2'): '*]*([)<>*()'})
    serialized = phantombuster.remoter.serialization.serialize_value(input_file)
    deserialized = phantombuster.remoter.serialization.deserialize_value(serialized, None)

    assert deserialized == input_file


def test_unit_regex_validation_1():
    regex1 = regex.compile("AAACC(?P<sample1>[CGT]{10})")
    regex2 = regex.compile("AAACC(?P<sample2>[ACGT]{10})")
    regex3 = regex.compile("AAACC(?P<lid>[ACGT]{10})")

    assert analysis.regexes_are_valid([regex1, regex2, regex3])


def test_unit_regex_validation_2():
    regex1 = regex.compile("G(?P<grna>[ACGT]{6})")
    regex2 = regex.compile("AACC")
    regex3 = regex.compile("AAACC(?P<lid>[ACGT]{10})")

    assert not analysis.regexes_are_valid([regex1, regex2, regex3])


def test_unit_plumbing_extract_read_information_1():
    read = b"D00689:314:CBYRBANXX:1:1104:4980:1967\t4\t*\t0\t0\t*\t*\t0\t0\tCCTATGGGCGTGGAAAGGACGAAACACCGCATGGCCCTGAAGAATGATGGTT\tBBBBBFFFFFFFFFFFFFBFFFFFFFFFFF<<FFFFFFF</F/<FFFFFFFF\tB2:Z:CGCTATCAA\tQ2:Z:<BBB<///B\tBC:Z:TTTCCCACCCT\tQT:Z:BBBBBFFFFFF\tRG:Z:CBYRBANXX.1\n"
    with tempfile.NamedTemporaryFile() as f:
        f.write(read)
        f.flush()

        b2regex = regex.compile("")
        bcregex = regex.compile("(?P<grna>[ACGT]{11})")
        queryregex = regex.compile("GGGC(?P<sample>[ACGT]{5})(?P<lid>[ACGT]{30})")

        regex_dictionary = {'B2': b2regex, 'BC': bcregex, 'query': queryregex}

        bamfile = iter(open_sequencing_file(f.name, type='sam'))
        read = next(bamfile)
        features = phantombuster.plumbing.extract_read(regex_dictionary, read)

        print(features)

        assert features["grna"] == "TTTCCCACCCT"
        assert features["sample"] == "GTGGA"
        assert features["lid"] == "AAGGACGAAACACCGCATGGCCCTGAAGAA"


def test_unit_plumbing_extract_read_information_2():
    read = b"D00689:314:CBYRBANXX:1:1104:4980:1967\t4\t*\t0\t0\t*\t*\t0\t0\tCCTATGGGCGTGGAAAGGACGAAACACCGCATGGCCCTGAAGAATGATGGTT\tBBBBBFFFFFFFFFFFFFBFFFFFFFFFFF<<FFFFFFF</F/<FFFFFFFF\tB2:Z:CGCTATCAA\tQ2:Z:<BBB<///B\tBC:Z:TTTCCCACCCT\tQT:Z:BBBBBFFFFFF\tRG:Z:CBYRBANXX.1\n"
    with tempfile.NamedTemporaryFile() as f:
        f.write(read)
        f.flush()

        b2regex = regex.compile("(?P<lid1>[ACGT]{9})")
        bcregex = regex.compile("(?P<grna>[ACGT]{11})")
        queryregex = regex.compile("GGGC(?P<sample>[ACGT]{5})(?P<lid2>[ACGT]{30})")
        regex_dictionary = {'B2': b2regex, 'BC': bcregex, 'query': queryregex}

        bamfile = iter(open_sequencing_file(f.name, type='sam'))
        read = next(bamfile)
        features = phantombuster.plumbing.extract_read(regex_dictionary, read)

        assert features["grna"] == "TTTCCCACCCT"
        assert features["sample"] == "GTGGA"
        assert features["lid"][:30] == "CGCTATCAAAAGGACGAAACACCGCATGGC"


def test_unit_plumbing_extract_read_information_uncommon_names():
    read = b"D00689:314:CBYRBANXX:1:1104:4980:1967\t4\t*\t0\t0\t*\t*\t0\t0\tCCTATGGGCGTGGAAAGGACGAAACACCGCATGGCCCTGAAGAATGATGGTT\tBBBBBFFFFFFFFFFFFFBFFFFFFFFFFF<<FFFFFFF</F/<FFFFFFFF\tB2:Z:CGCTATCAA\tQ2:Z:<BBB<///B\tBC:Z:TTTCCCACCCT\tQT:Z:BBBBBFFFFFF\tRG:Z:CBYRBANXX.1\n"
    with tempfile.NamedTemporaryFile() as f:
        f.write(read)
        f.flush()

        b2regex = regex.compile("(?P<cellidentifier1>[ACGT]{9})")
        bcregex = regex.compile("(?P<cellline>[ACGT]{11})")
        queryregex = regex.compile("GGGC(?P<organoid>[ACGT]{5})(?P<cellidentifier2>[ACGT]{30})")
        regex_dictionary = {'B2': b2regex, 'BC': bcregex, 'query': queryregex}

        bamfile = iter(open_sequencing_file(f.name, type='sam'))
        read = next(bamfile)
        features = phantombuster.plumbing.extract_read(regex_dictionary, read)

        assert features["cellline"] == "TTTCCCACCCT"
        assert features["organoid"] == "GTGGA"
        assert features["cellidentifier"][:30] == "CGCTATCAAAAGGACGAAACACCGCATGGC"


def test_unit_plumbing_extract_read_information_as_lt47_experiment():
    read = b"D00689:314:CBYRBANXX:1:1104:13784:1992\t4\t*\t0\t0\t*\t*\t0\t0\tACGGGTGAAGTGGAAAGGACGAAACACCGGTACAAGATCAACAACAAAGGTT\tBB//BFBFF/FFFFFFFFFFFFFFFFFFB<FF<FFF///<FFFF</FBFFFF\tB2:Z:CGCTATCAA\tQ2:Z:<BBBB/F/F\tBC:Z:CGTATTTTAGT\tQT:Z:B/<<<FFFFFF\tRG:Z:CBYRBANXX.1"
    with tempfile.NamedTemporaryFile() as f:
        f.write(read)
        f.flush()

        b2regex = regex.compile("^[ACGTN]{3}(?P<sample>[ACGTN]{5})")
        bcregex = regex.compile("")
        queryregex = regex.compile("(?P<lid>[ACGTN]{5,6}(?P<lib>ACGT|GTAC){s<=1}[ACGTN]+)")
        regex_dictionary = {'B2': b2regex, 'BC': bcregex, 'query': queryregex}

        bamfile = iter(open_sequencing_file(f.name, type='sam'))
        read = next(bamfile)
        features = phantombuster.plumbing.extract_read(regex_dictionary, read)
        print(features)

        assert features["sample"] == "TATCA"
        assert features["lid"] == "CGGGTGAAGTGGAAAGGACGAAACACCGGTACAAGATCAACAACAAAGGTT"
        assert features["lib"] == "AAGT"


def test_unit_table_to_counter():
    table1 = pl.DataFrame({"lid":        [    "AAC",     "ACG",     "GTA"],
                           "samplename": ["sample1", "sample1", "sample1"],
                           "lib":        [   "lib1",    "lib1",    "lib2"],
                           "reads":      [        1,         1,         5],
                          })

    c = table_to_counter(table1, ["lid", "samplename", "lib"])
    print(c)
    assert c[("AAC", "sample1", "lib1")] == 1
    assert c[("ACG", "sample1", "lib1")] == 1
    assert c[("GTA", "sample1", "lib2")] == 5


def test_unit_error_corrector_1():
    ec = ErrorCorrector(4, 0, 0, 1)
    tags = {Umi(0, 0, "AAAA"): UmiCount(1, 1), Umi(0, 0, "AAAT"): UmiCount(10,1), Umi(0, 0, "AATA"): UmiCount(10, 1), Umi(0,0, "ATAA"): UmiCount(10, 1)}
    corr_lids = ec.process(tags)
    # AAAA can be corrected to AAAT, AATA or ATAA and is thus ambigious and should be removed.
    assert Umi(0, 0, "AAAA") not in corr_lids[1]
    assert corr_lids[0][Umi(0,0, "AAAT")].count == 10
    assert corr_lids[0][Umi(0,0, "AATA")].count == 10
    assert corr_lids[0][Umi(0,0, "ATAA")].count == 10
    assert len(corr_lids[0]) == 3


def test_unit_error_corrector_2():
    ec = ErrorCorrector(50, 0, 0, 1)
    tags = {Umi(0, 0, "AAAAGTACGTGTCAGACTGTGAGAGACTGACAGTCAGAGTCTGACAACGT"): UmiCount(112, 1), Umi(0, 0, "AAAAGTACGTGTCAGACTGTGAGAGACTAACAGTCAGAGTCTGACAACGT"): UmiCount(2,1), Umi(0, 0, "AAAAGTACGTGAGAGACTCTGTGTCACTCAGTCTGACAGTCTGTGAACGT"): UmiCount(6, 1), Umi(0,0, "AAAAGTACGTGAGAGACTCCGTGTCACTCAGTCTGACAGTCTGTGAACGT"): UmiCount(2, 1), Umi(0, 0, "AAAAGTACGTGTCAGACTGTGAGAGACTGACAGTCAGAGTCTGGCAACGT"): UmiCount(2, 1), Umi(0,0,"AAAAGTACGTGTCAGACTGTGAGAGACTGACAGTCAGAGTCTTACAACGT"): UmiCount(6, 1), Umi(0,0,"AAAAGTACGTGTGACTCTGTGTGTGTGACTGTGTCTCTGTCTCACAACGT"): UmiCount(40, 1)}
    corr_lids = ec.process(tags)

    for key, count in corr_lids[0].items():
        print(key)
        print(count)
    for lid, correct in corr_lids[1].items():
        print(f"{lid.tag} {corr_lids[0][lid]} -> {correct.tag} : {corr_lids[0][correct]}")
    assert corr_lids[0][Umi(start=0, end=0, tag='AAAAGTACGTGTCAGACTGTGAGAGACTGACAGTCAGAGTCTGACAACGT')].count == 122
    assert corr_lids[0][Umi(start=0, end=0, tag='AAAAGTACGTGAGAGACTCTGTGTCACTCAGTCTGACAGTCTGTGAACGT')].count == 8
    assert corr_lids[0][Umi(start=0, end=0, tag='AAAAGTACGTGTGACTCTGTGTGTGTGACTGTGTCTCTGTCTCACAACGT')].count == 40

    assert sum(tag.count for tag in corr_lids[0].values()) == 170


def test_unit_plumbing_error_correct_1():
    sample = [ "S1",    "S1",    "S1",    "S2",    "S2",    "S2"]
    tags = ["AAAAC", "AAACC", "AACCC", "ACCCC", "CCCCC", "GTAAC"]
    grna = [    "A",     "A",     "A",     "A",     "A",     "A"]
    reads = [     1,       2,     10 ,      1 ,     10 ,      1]
    table = pl.DataFrame({"sample": sample, "tags": tags, "reads": reads, "grna": grna})
    barcode_hierarchy = [{"name": "sample", "type": "reference"}, {"name": "grna", "type": "reference"}, {"name": "tags", "type": "random"}]
    corrected_table = plumbing.error_correct(table, barcode_hierarchy, 1).to_dict()
    assert (corrected_table["sample"] == ["S1", "S2", "S2"]).all()
    assert (corrected_table["tags"]  == ["AACCC", "CCCCC", "GTAAC"]).all()
    assert (corrected_table["reads"] == [     13,      11,       1]).all()
    assert (corrected_table["grna"] == ["A", "A", "A"]).all()


def test_unit_plumbing_error_correct_2():
    tags = ["AAAAC", "AAACC", "AACCC", "ATTTT", "TTTTT", "GTAAC"]
    reads = [     1,       2,     10 ,      1 ,     10 ,      1]
    table = pl.DataFrame({"tags": tags, "reads": reads})
    barcode_hierarchy = [{"name": "tags", "type": "random"}]
    corrected_table = plumbing.error_correct(table, barcode_hierarchy, 1).to_dict()
    assert (corrected_table["tags"]  == ["AACCC", "GTAAC", "TTTTT"]).all()
    assert (corrected_table["reads"] == [     13,        1,       11]).all()


def test_unit_plumbing_error_correct_3_first_base():
    sample = [ "S1",    "S1",    "S1"]
    tags = ["AAAAA", "AAACC", "CAAAA"]
    grna = [    "A",     "A",     "A"]
    reads = [    10,       10,     1 ]
    table = pl.DataFrame({"sample": sample, "tags": tags, "reads": reads, "grna": grna})
    barcode_hierarchy = [{"name": "sample", "type": "reference"}, {"name": "grna", "type": "reference"}, {"name": "tags", "type": "random"}]
    corrected_table = plumbing.error_correct(table, barcode_hierarchy, 1).to_dict()
    assert (corrected_table["sample"] == ["S1", "S1"]).all()
    assert (corrected_table["tags"]  == ["AAAAA", "AAACC"]).all()
    assert (corrected_table["reads"] == [     11,      10]).all()
    assert (corrected_table["grna"] == ["A", "A"]).all()


def test_unit_plumbing_apply_correction():
    tags = ["AAAAC", "AAACC", "AACCC", "ACCCC", "CCCCC", "GTAAC"]
    reads = [     1,       2,     10 ,      1 ,     10 ,      1]
    corrections = {"AAAAC": "AACCC", "AAACC": "AACCC", "AACCC": "AACCC", "CCCCC": "CCCCC", "GTAAC": "GTAAC"}
    table = pl.DataFrame({"tags": tags, "reads": reads})
    table = plumbing.apply_correction(table, "tags", corrections)
    table = table.sort("tags")
    d = table.to_dict()
    assert d["tags"].to_list()  == ["AACCC", "CCCCC", "GTAAC"]
    assert d["reads"].to_list() == [     13,      10,       1]


def test_unit_plumbing_sort_and_partition():
    table = pl.DataFrame({
        'sample': ["A", "A", 'A', 'A', 'B', 'B', "C", "C"],
        'reads' : [ 3 ,  1 ,  1 ,  1,   2 ,  3 ,  4 ,  1 ],
        'lib'   : ["a", "a", 'a', 'b', 'a', 'a', "a", "b"],
        })

    partitions = plumbing.sort_and_partition(table, ('sample', 'lib'), ('sample', 'lib'))
    if len(table) != sum(len(p) for p in partitions.values()):
        print(f"Total table length is {len(table)}")
        for p in partitions.values():
            print(f"Partition length {len(p)}")
    assert len(table) == sum(len(p) for p in partitions.values())
    assert ('A', 'a') in partitions
    assert ('A', 'b') in partitions
    assert ('B', 'a') in partitions
    assert ('B', 'b') not in partitions

    assert np.sum(partitions[('A', 'a')]['reads'].to_numpy()) == 5
    assert np.sum(partitions[('A', 'b')]['reads'].to_numpy()) == 1
    assert np.sum(partitions[('B', 'a')]['reads'].to_numpy()) == 5
    assert ("B", "b") not in partitions
    assert np.sum(partitions[('C', 'a')]['reads'].to_numpy()) == 4
    assert np.sum(partitions[('C', 'b')]['reads'].to_numpy()) == 1

    assert len(partitions[('A', 'a')]['reads'].to_numpy()) == 3
    assert len(partitions[('A', 'b')]['reads'].to_numpy()) == 1
    assert len(partitions[('B', 'a')]['reads'].to_numpy()) == 2
    assert len(partitions[('C', 'a')]['reads'].to_numpy()) == 1
    assert len(partitions[('C', 'b')]['reads'].to_numpy()) == 1

def test_unit_extract_section():

    fixed = "GGGC"

    b2regex = "(?P<umi>[ACGT]{9})"
    bcregex = "(?P<grna>[ACGT]{11})"
    queryregex = "(?P<lib>GGGC)(?P<sample>[ACGT]{5})(?P<lid>[ACGT]{30})"
    regex_dictionary = phantombuster.config_files.RegexDictionary()
    regex_dictionary.add_regex('B2', b2regex)
    regex_dictionary.add_regex('BC', bcregex)
    regex_dictionary.add_regex('query', queryregex)

    samples = ["GTGGA", "TTTGA"]
    sample_dict = {s: f"Sample{i}" for i, s in enumerate(samples)}

    samples0_mutated = "GGGGA"

    genes = ["TTTCCCACCCT"]
    gene_dict = {s: f"Gene{i}" for i, s in enumerate(genes)}

    lib_dict = {"GGGC": "lib1"}

    lid1 = "AAGGACGAAACACCGCATGGCCCTGAAGAA"
    lid2 = "CAGGACGAAACACCGCATGGCCCTGAAGAA" # hamming distance 1 to lid1

    umi1 = "CGCTATCAA"
    umi2 = "AAGTATCAA"
    umi3 = "CGCTAGCCT"

    reads = [
            make_read(f"{fixed}{samples[0]}{lid1}", umi1, genes[0]),
            make_read(f"{fixed}{samples0_mutated}{lid1}", umi1, genes[0]), # should be collapsed with read 1
            make_read(f"{fixed}{samples[0]}{lid1}", umi1, genes[0]),
            make_read(f"{fixed}{samples[0]}{lid1}", umi2, genes[0]),
            make_read(f"{fixed}{samples[0]}{lid2}", umi1, genes[0]),
            make_read(f"{fixed}{samples[0]}{lid2}", umi3, genes[0]),
            make_read(f"{fixed}{samples[1]}{lid2}", umi3, genes[0]),
            ]


    barcode_hierarchy = [{"name": "lib", "type": "reference", "reference": lib_dict, "referencefile": None, "threshold": 0, "min_length": None, "max_length": None},
                         {"name": "sample", "type": "reference", "reference": sample_dict, "threshold": 1, "min_length": None, "max_length": None},
                         {"name": "grna", "type": "reference", "reference": gene_dict, "threshold": 0, "min_length": None, "max_length": None},
                         {"name": "lid", "type": "auto", "threshold": "auto", "min_length": None, "max_length": None},
                         {"name": "umi", "type": "auto", "threshold": "auto", "min_length": None, "max_length": None},
                         ]


    with tempfile.NamedTemporaryFile(suffix='.sam') as f:
        for read in reads:
            f.write(read)
        f.flush()

        lidonly = False
        count_qc = False

        input_file = phantombuster.config_files.InputFile(f.name, '', '')

        f = phantombuster.io_.SamSequencingFile(input_file.path)
        assert len(list(f)) == len(reads)

        input_group = phantombuster.config_files.InputGroup([f], "")

        table, stats = phantombuster.plumbing.deduplicate_section(input_group, None, regex_dictionary, barcode_hierarchy, count_qc)

    cols = table.columns
    cols.remove("reads")

    c = table_to_counter(table, cols)

    assert sum(c.values()) == 7

    for k, v in c.items():
        print(k, v)

    assert c[("lib1", "Sample0", "Gene0", lid1, umi1)] == 3
    assert c[("lib1", "Sample0", "Gene0", lid1, umi2)] == 1
    assert c[("lib1", "Sample0", "Gene0", lid2, umi1)] == 1
    assert c[("lib1", "Sample0", "Gene0", lid2, umi3)] == 1
    assert c[("lib1", "Sample1", "Gene0", lid2, umi3)] == 1

    assert sum(c.values()) == len(reads)


def test_unit_deduplication_persister():

    from phantombuster import stores
    from phantombuster.remoter.task import Task

    result = [{("lid1", "lib1"): 1, ("lid2", "lib1"): 2}, [3, {"sample": {"sample1": 3}}, {"sample":0} ]]

    result[0] = deduplicator_to_table(Counter(result[0]), [{'name': "lid", 'type': 'random'}, {'name': "lib", 'type': 'reference'}])

    with tempfile.TemporaryDirectory() as tmpdirname:
        id = stores.save(result, dir=tmpdirname)
        v = stores.load(id, dir=tmpdirname)
        v = (v[0].collect(), v[1])

    r1 = table_to_counter(v[0], ["lid", "lib"])
    r2 = table_to_counter(v[0], ["lid", "lib"])

    assert r1[("lid1", "lib1")] == r2[("lid1", "lib1")]
    assert r1[("lid2", "lib1")] == r2[("lid2", "lib1")]

    assert r1 == r2


def test_unit_combine():

    fixed = "GGGC"

    b2regex = "(?P<umi>[ACGT]{9})"
    bcregex = "(?P<grna>[ACGT]{11})"
    queryregex = "(?P<lib>GGGC)(?P<sample>[ACGT]{5})(?P<lid>[ACGT]{30})"
    regex_dictionary = phantombuster.config_files.RegexDictionary()
    regex_dictionary.add_regex('B2', b2regex)
    regex_dictionary.add_regex('BC', bcregex)
    regex_dictionary.add_regex('query', queryregex)

    samples = ["GTGGA"]
    sample_dict = {s: f"Sample{i}" for i, s in enumerate(samples)}

    genes = ["TTTCCCACCCT"]
    gene_dict = {s: f"Gene{i}" for i, s in enumerate(genes)}

    lib_dict = {"GGGC": "lib1"}

    lid1 = "AAGGACGAAACACCGCATGGCCCTGAAGAA"
    lid2 = "CAGGACGAAACACCGCATGGCCCTGAAGAA"

    umi1 = "CGCTATCAA"
    umi2 = "AAGTATCAA"
    umi3 = "CGCTAGCCT"

    reads = [
            make_read(f"{fixed}{samples[0]}{lid1}", umi1, genes[0]),
            make_read(f"{fixed}{samples[0]}{lid1}", umi1, genes[0]),
            make_read(f"{fixed}{samples[0]}{lid1}", umi2, genes[0]),
            make_read(f"{fixed}{samples[0]}{lid2}", umi1, genes[0]),
            make_read(f"{fixed}{samples[0]}{lid2}", umi3, genes[0]),
            ]

    #                    error is at line in read3                                         |


    barcode_hierarchy = [{"name": "lib", "type": "reference", "reference": lib_dict, "referencefile": None, "threshold": 0, "min_length": None, "max_length": None},
                         {"name": "sample", "type": "reference", "reference": sample_dict, "threshold": 0, "min_length": None, "max_length": None},
                         {"name": "grna", "type": "reference", "reference": gene_dict, "threshold": 0, "min_length": None, "max_length": None},
                         {"name": "lid", "type": "auto", "threshold": "auto", "min_length": None, "max_length": None},
                         {"name": "umi", "type": "auto", "threshold": "auto", "min_length": None, "max_length": None},
                         ]

    with tempfile.NamedTemporaryFile(suffix='.sam') as f:
        for read in reads:
            f.write(read)
        f.flush()

        lidonly = False
        count_qc = False

        input_file = phantombuster.config_files.InputFile(f.name, '', '')

        f = phantombuster.io_.SamSequencingFile(input_file.path)
        assert len(list(f)) == len(reads)

        input_group = phantombuster.config_files.InputGroup([f], "")

        out1 = phantombuster.plumbing.deduplicate_section(input_group, None, regex_dictionary, barcode_hierarchy, count_qc)
        out2 = phantombuster.plumbing.deduplicate_section(input_group, None, regex_dictionary, barcode_hierarchy, count_qc)

    table, stats = phantombuster.plumbing.combine([out1, out2], barcode_hierarchy)

    cols = table.columns
    cols.remove("reads")

    c = table_to_counter(table, cols)

    for k, v in c.items():
        print(k, v)

    assert c[("lib1", "Sample0", "Gene0", lid1, umi1)] == 4
    assert c[("lib1", "Sample0", "Gene0", lid1, umi2)] == 2
    assert c[("lib1", "Sample0", "Gene0", lid2, umi1)] == 2
    assert c[("lib1", "Sample0", "Gene0", lid2, umi3)] == 2

    assert sum(c.values()) == 2*len(reads)


def test_unit_combine_plumbing():

    table1 = pl.DataFrame({"lid":    [    "AAC",     "ACG",     "GTA"],
                           "sample": ["sample1", "sample1", "sample1"],
                           "reads":  [       1 ,        1 ,        5 ],
                          })

    table2 = pl.DataFrame({"lid":    [    "AAC",     "ACG",     "GTA"],
                           "sample": ["sample1", "sample1", "sample2"],
                           "reads":  [       1 ,        1 ,        5 ],
                          })

    bhc = [{"name": "sample", "type": "reference"}, {"name": "lid", "type": "random"}]
    combined = phantombuster.plumbing._combine_duplicators_polars(table1, table2, bhc)


    print(combined)
    assert len(combined) == 4
    assert combined["sample"][3] == "sample2"



def test_unit_plumbing_correction():
    tags  = ["AAAA", "AAAT", "ACGT", "AATT", "AACC", "AACT"]
    reads = [     32,      8,      4,      2,    16 ,     1]

    corrections = plumbing.calculate_corrections(tags, reads, 4, 1)

    assert corrections["AAAA"] == "AAAA"
    assert corrections["AAAT"] == "AAAA"
    assert corrections["ACGT"] == "ACGT"
    assert corrections["AATT"] == "AAAA"
    assert "AACT" not in corrections
    assert corrections["ACGT"] == "ACGT"
    assert len(corrections) == 5


def test_unit_calculate_correction():
    lids = ['AAACCCGGGTTT', 'GAACCCGGGTTT', 'GGACCCGGGTTT', 'GGGCCCGGGTTT', 'GGGTTTAAACCC']
    reads = [10, 3, 1, 10, 11]

    corrections = plumbing.calculate_corrections(np.array(lids), np.array(reads), len(lids[0]), 1)

    assert len(corrections) == 4
    assert corrections['AAACCCGGGTTT'] == 'AAACCCGGGTTT'
    assert corrections['GAACCCGGGTTT'] == 'AAACCCGGGTTT'
    assert corrections['GGGCCCGGGTTT'] == 'GGGCCCGGGTTT'
    assert corrections['GGGTTTAAACCC'] == 'GGGTTTAAACCC'
    assert corrections.get('GGACCCGGGTTT', None) == None


def test_unit_deduplicate_single():
    sample = ['s1', 's1', 's1', 's2', 's2']
    lids =   [ 'A',  'A',  'A',  'A',  'B']
    umis =   [ 'X',  'X',  'Y',  'X',  'X']
    reads =  [  1 ,   2 ,   1 ,   2,    5 ]


    sample, lids, umis, reads = np.array(sample), np.array(lids), np.array(umis), np.array(reads)

    table = pl.DataFrame({"sample": sample, "lid": lids, "umi": umis, "reads": reads})
    t = plumbing.deduplicate_single(table)

    c = {col: t[col].to_numpy() for col in t.columns}

    assert np.sum(c['reads']) == np.sum(reads)
    assert len(c['reads']) == 4


def test_unit_remove_ambigious():
    sample = [ 's1',  's2']
    lids =   ['ACG', 'NNN']
    reads =  [   3 ,    2 ]

    barcode_list= [{"name": "sample", "type": "reference", "reference": None, "referencefile": None, "threshold": 0, "min_length": None, "max_length": None},
                   {"name": "lid", "type": "random", "threshold": "auto", "min_length": None, "max_length": None},
                  ]

    sample, lids, reads = np.array(sample), np.array(lids), np.array(reads)

    table = pl.DataFrame({"sample": sample, "lid": lids, "reads": reads})
    table = plumbing.remove_ambigious(table, barcode_list)

    assert table['reads'].sum() == 3
    assert len(table) == 1


def test_unit_call_hopping():

    sample =     [ 's1',  's2']
    lids =       ['ACG', 'NNN']
    reads =      [   3 ,    2 ]
    threshold1 = [   2 ,    3 ]
    threshold2 = [   1 ,    1 ]

    table = df = pl.DataFrame({"sample": sample, "lids": lids, "reads": reads})
    th1 = pl.DataFrame({"threshold": threshold1})
    th2 = pl.DataFrame({"threshold": threshold1})

    df = plumbing.call_hopping(df, [th1, th2])

    assert len(df) == 1
    assert df.columns == table.columns


@pytest.mark.slow
def test_unit_combine_large_tables():
    bc_hierarchy = [{"name": "LID", "type": "random"}]
    #N = 10000000
    N = 10000000

    a = pl.select(pl.arange(N).cast(str).alias("LID"), pl.lit(1).alias("reads"))
    a = a.sort("LID")
    d = phantombuster.plumbing._combine_duplicators_polars(a, a, bc_hierarchy)
    assert len(d) == N


def test_unit_statistics():
    df = pl.DataFrame({'grna': ['A', 'A', 'A', 'B', 'C', 'C'],
                       'category': ['control', 'control', 'control', 'test', 'test', 'test'],
                       'reads': [10, 20, 30, 25, 5, 5]})

    r = phantombuster.statistics.calculate_pvalues(df, column='reads', group_by=['grna'], N=10000)
    for row in r.iter_rows(named=True):
        if row['grna'] == 'B':
            assert row['pvalue'] - 0.66 < 0.05
            assert row['lineages'] == 1
        elif row['grna'] == 'C':
            assert row['pvalue'] == 0.0
            assert row['lineages'] == 2


# -- Integration test helpers -- #


def prepare_dir(template):
    # make temporary directory for all data
    tmpdir = pjoin("/tmp", uuid.uuid4().hex)
    os.mkdir(tmpdir)

    template_dir = pjoin(os.path.dirname(os.path.abspath(__file__)), 'test_data_files', template)
    files = glob.glob(pjoin(template_dir, '*'))

    for p in files:
        p = os.path.basename(p)
        copy_file_efficiently(pjoin(template_dir, p), pjoin(tmpdir, p))

    return tmpdir


def execute_in(tmpdir):
    old_wd = os.getcwd()
    os.chdir(tmpdir)
    yield 'input_files.csv', 'regexes.csv'
    os.chdir(old_wd)


# -- Integration Test Fixtures -- #

@pytest.fixture()
def setup_lineage_data_small():
    tmp_dir = prepare_dir('small')
    yield from execute_in(tmp_dir)


@pytest.fixture()
def setup_lineage_data_small_fastq():
    tmp_dir = prepare_dir('small_fastq')
    yield from execute_in(tmp_dir)


@pytest.fixture()
def setup_lineage_data_medium():
    tmp_dir = prepare_dir('medium')
    yield from execute_in(tmp_dir)


@pytest.fixture()
def setup_lineage_data_big():
    tmp_dir = prepare_dir('big')
    yield from execute_in(tmp_dir)


@pytest.fixture()
def setup_lineage_data_1010():
    tmp_dir = prepare_dir('1010')
    yield from execute_in(tmp_dir)



# -- Integration Tests -- #


def test_integration_read_regex_file(setup_lineage_data_small):
    regex_dict = phantombuster.config_files.read_regex_file('regexes.csv')
    regexes = regex_dict.get_regexes_for_group('*')
    assert regexes['B2'] == regex.compile('^[ACGTN]{3}(?P<sample>[ACGTN]{5})')
    assert regexes['query'] == regex.compile('(?P<lid>[ACGTN]{5,6}(?P<lib>ACGT|GTAC){s<=1}[ACGTN]+)')

    assert regexes['B2'].match('AAACGTAGC').group('sample') == 'CGTAG'
    assert regexes['query'].match('AAAAAGTACAAAAAAAAAAA').group('lib') == 'GTAC'


def test_integration_read_input_files_file(setup_lineage_data_small):
    input_groups = phantombuster.config_files.read_input_files_file('input_files.csv')
    assert phantombuster.config_files.InputFile('101.bam', '101.bam', '') in input_groups[0].files


def test_integration_read_barcode_hierarchy(setup_lineage_data_small):
    bch = phantombuster.config_files.read_barcode_hierarchy_file('barcode_hierarchy.csv')
    assert [bc["name"] for bc in bch] == ['lib', 'sample', 'lid']
    assert [bc["type"] for bc in bch] == ['reference', 'reference', 'random']


def test_integration_indexing_bam(setup_lineage_data_1010):
    bamfiles, regex_file = setup_lineage_data_1010
    idx = index_bamfile("1010.bam", 1)
    every_read_index = []
    f = open_sequencing_file('1010.bam')

    for read in f:
        every_read_index.append(f.tell())

    with_index = []
    for start, end in pairwise(idx):
        with open_sequencing_file('1010.bam') as f:
            f.seek(start)
            for read in f:
                with_index.append(f.tell())
                if f.tell() >= end:
                    break

    assert with_index == every_read_index


def test_integration_indexing_fastq(setup_lineage_data_1010):
    bamfiles, regex_file = setup_lineage_data_1010
    idx = index_fastqgz_file("1010.fastq.gz", 1)
    with gzip.open("1010.fastq.gz") as f:
        lines = f.readlines()

    every_read_index = []
    i = 0
    index = 0
    while True:
        every_read_index.append(index)
        if i+4 < len(lines):
            index += sum(len(l) for l in lines[i:i+4])
            i += 4
        else:
            break

    assert idx[:-1] == every_read_index
   

from phantombuster.io_ import open_sequencing_file

def test_integration_read_fastq(setup_lineage_data_1010):
    bamfiles, regex_file = setup_lineage_data_1010


    with open_sequencing_file('1010.fastq.gz') as f:
        fi = iter(f)
        read = next(fi)
        assert read['name'] == '@D00689:314:CBYRBANXX:1:1104:13784:1992'
        assert read['seq'] == 'ACGGGTGAAGTGGAAAGGACGACACACCGGTACAAGATCAACAACAAAGGTT'


def test_integration_read_bam(setup_lineage_data_1010):
    bamfiles, regex_file = setup_lineage_data_1010

    with open_sequencing_file('1010.bam') as f:
        fi = iter(f)
        read = next(fi)
        assert read['query'] == 'ACGGGTGAAGTGGAAAGGACGACACACCGGTACAAGATCAACAACAAAGGTT'
        assert read['B2'] == 'CGCTATCAA'


@pytest.mark.slow
def test_integration_fullrun_only(setup_lineage_data_small):
    input_files_file, regex_file = setup_lineage_data_small
    project = Project("out")

    core.demultiplex(input_files_file, regex_file=regex_file, barcode_hierarchy_file="barcode_hierarchy.csv", project=project, debug=True, show_qc=False)

    result = read_file(pjoin("out", 'data', "demultiplex.parquet")).collect()
    r1 = result.filter(pl.col("lid") == "CGGGTGAAGTGGAAAGGACGAAACACCGGTACAAGATCAACAACAAAGGT")
    r2 = result.filter(pl.col("lid") == "CGGGTGAAGTGGAAAGGACGACACACCGGTACAAGATCAACAACAAAGGT")

    assert r1["lib"][0] == "lib1"
    assert r1["reads"][0] == 100

    assert r2["lib"][0] == "lib1"
    assert r2["reads"][0] == 1


@pytest.mark.slow
def test_integration_fullrun_fastq_only(setup_lineage_data_small_fastq):
    input_files_file, regex_file = setup_lineage_data_small_fastq

    project = Project("out")

    core.demultiplex(input_files_file, regex_file=regex_file, barcode_hierarchy_file="barcode_hierarchy.csv", project=project, debug=True, show_qc=False)

    result = read_file(pjoin("out", 'data', "demultiplex.parquet")).collect()
    r1 = result.filter(pl.col("lid") == "CGGGTGAAGTGGAAAGGACGAAACACCGGTACAAGATCAACAACAAAGGT")
    r2 = result.filter(pl.col("lid") == "CGGGTGAAGTGGAAAGGACGACACACCGGTACAAGATCAACAACAAAGGT")

    assert r1["lib"][0] == "lib1"
    assert r1["reads"][0] == 100

    assert r2["lib"][0] == "lib1"
    assert r2["reads"][0] == 1


@pytest.mark.slow
def test_integration_fullrun_many_only(setup_lineage_data_medium):
    input_files_file, regex_file = setup_lineage_data_medium
    project = Project("out")

    core.demultiplex(input_files_file, regex_file=regex_file, barcode_hierarchy_file="barcode_hierarchy.csv", project=project, debug=True, show_qc=False)

    result = read_file(pjoin("out", 'data', "demultiplex.parquet")).collect()
    print(result)
    assert result.filter(pl.col("lid") == "GGGCGCCGAAAGGGACATTCCAGCAGCTCT")["lib"][0] == "lib1"
    assert result.filter(pl.col("lid") == "GGGCGCCGAAAGGGACATTCCAGCAGCTCT")["reads"][0] == 1025
    assert result["reads"].sum() == 10000


@pytest.mark.slow
def test_integration_fullrun_no_prefix(setup_lineage_data_small):
    input_files_file, regex_file = setup_lineage_data_small

    project = Project("out")

    core.demultiplex(input_files_file, regex_file=regex_file, barcode_hierarchy_file="barcode_hierarchy.csv", project=project, debug=True, show_qc=False)

    result = read_file(pjoin("out", 'data', "demultiplex.parquet")).collect()
    r1 = result.filter(pl.col("lid") == "CGGGTGAAGTGGAAAGGACGAAACACCGGTACAAGATCAACAACAAAGGT")
    r2 = result.filter(pl.col("lid") == "CGGGTGAAGTGGAAAGGACGACACACCGGTACAAGATCAACAACAAAGGT")
    assert r1["lib"][0] == "lib1"
    assert r1["reads"][0] == 100

    assert r2["lib"][0] == "lib1"
    assert r2["reads"][0] == 1


@pytest.mark.slow
def test_integration_fullrun_multiple_files(setup_lineage_data_big):
    input_files_file, regex_file = setup_lineage_data_big
    project = Project("out")

    core.demultiplex(input_files_file, regex_file=regex_file, barcode_hierarchy_file="barcode_hierarchy.csv", project=project, debug=True, show_qc=False)

    result = read_file(pjoin("out", 'data', "demultiplex.parquet")).collect()
    for data in result.iter_rows():
        print(data)

    r = result.filter(pl.col("lid") == "CGGGTGAAGTGGAAAGGACGAAACACCGGTACAAGATCAACAACAAAGGT")

    print(r)
    assert r["lib"][0] == "lib1"
    assert r["reads"][0] == 500

@pytest.mark.slow
def test_integration_pvalue(setup_lineage_data_big):
    input_files_file, regex_file = setup_lineage_data_big
    project = Project("out")

    core.demultiplex(input_files_file, regex_file=regex_file, barcode_hierarchy_file="barcode_hierarchy.csv", project=project, debug=True, show_qc=False)
    core.error_correct(project, error_threshold=1, barcode_hierarchy_file="barcode_hierarchy.csv", remove_ambigious=True)
    core.hopping_removal(project, ['lid'], 0.05)
    core.threshold(project, 'thresholds.csv')
    core.pvalue(project, project.threshold_output_path, 'categories.csv', group_by=['sample'])


@pytest.mark.slow
def test_integration_fullrun_with_ec(setup_lineage_data_big):
    input_files_file, regex_file = setup_lineage_data_big
    project = Project("out")

    core.demultiplex(input_files_file, regex_file=regex_file, barcode_hierarchy_file="barcode_hierarchy.csv", project=project, debug=True, show_qc=False)

    print("Finished phantombuster demultiplex")

    print("E16: All files in temporary directory")
    for root, dirs, files in os.walk("."):
        print(root)
        for f in files:
            print(pjoin(root, f))

    core.error_correct(project=project, error_threshold=1, barcode_hierarchy_file='barcode_hierarchy.csv', remove_ambigious=True)

    print("E15: All files in temporary directory")
    for root, dirs, files in os.walk("."):
        print(root)
        for f in files:
            print(pjoin(root, f))

    before_ec = read_file(pjoin("out", 'data', "demultiplex.parquet")).collect()

    for lib, sample, lid, reads in zip(before_ec["lib"].to_list(), before_ec["sample"].to_list(), before_ec["lid"].to_list(), before_ec["reads"].to_list()):
        print(lib, sample, lid, reads)

    df = read_file(pjoin("out", 'data', "error_correct.parquet")).collect()

                                               # ERROR ->@<-# 
                                  #'CGGGTGAAGTGGAAAGGACGACACACCGGTACAAGATCAACAACAAAGGT'
    r = df.filter(pl.col('lid') == "CGGGTGAAGTGGAAAGGACGAAACACCGGTACAAGATCAACAACAAAGGT")

    assert r["lib"][0] == "lib1"
    assert r["reads"][0] == 508

    assert len(df['sample'].unique()) == 2

    for column_name in df.columns:
        print(column_name, df[column_name].dtype)

    #subprocess.run(["phantombuster", "--verbose", "hopping_removal", "--outdir=out"], timeout=20, check=True)


    print("E17 All files in temporary directory")
    for root, dirs, files in os.walk("."):
        print(root)
        for f in files:
            print(pjoin(root, f))

    #result = pa.parquet.read_table(pjoin("out", "data", "hopping_removal", "25-H9-day00-01.parquet")).to_pandas()

@pytest.mark.slow
def test_integration_logger_handler():

    zmq_log_handler = None

    try:
        logger = logging.getLogger()
        #configure_logging(outputlog="worker" + name, verbose=True)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(process)s - %(levelname)s - %(message)s")
        logger.setLevel(logging.DEBUG)

        context = zmq.Context()

        zmq_log_handler = PushHandler("inproc://test_logger", context=context)
        zmq_log_handler.setLevel(logging.DEBUG)
        logger.addHandler(zmq_log_handler)

        thread = threading.Thread(target=logging.debug, args=["hello"])
        thread.start()

        socket = context.socket(zmq.PULL)
        socket.bind("inproc://test_logger")
        msg = None
        count = 0
        while msg is None and count <= 10:
            time.sleep(1)
            try:
                msg = socket.recv(flags=zmq.NOBLOCK)
            except zmq.error.Again as e:
                count += 1
        if msg is None:
            msg = socket.recv(flags=zmq.NOBLOCK)

        assert b"hello" in msg
    finally:
        if zmq_log_handler:
            logger.removeHandler(zmq_log_handler)



"""
Algorithmic steps, functions take tables, scalars, etc. as input and also output tables, scalars etc.
Does not deal with IO. Functions need to be stitched together to do something useful.
"""
import time
import logging
import pysam
from collections import Counter, defaultdict, deque
import functools
import regex
from typing import Optional, Dict
from phantombuster.error_corrector import ErrorCorrector, Umi, UmiCount
from phantombuster.io_ import open_sequencing_file
import numpy as np
from itertools import tee
import scipy.stats
from dataclasses import dataclass
from scipy.stats._discrete_distns import binom
import polars as pl

from phantombuster import bamindexer

def collect(df):
    if hasattr(df, "collect"):
        return df.collect()
    return df

def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)

_marker = object()

try:
    StringCounter = Counter[str] # type: ignore
except:
    StringCounter = Counter # type: ignore

featuretype = tuple

# copied from more-itertools https://github.com/more-itertools/more-itertools 
class peekable:
    """Wrap an iterator to allow lookahead and prepending elements.

    Call :meth:`peek` on the result to get the value that will be returned
    by :func:`next`. This won't advance the iterator:

        >>> p = peekable(['a', 'b'])
        >>> p.peek()
        'a'
        >>> next(p)
        'a'

    Pass :meth:`peek` a default value to return that instead of raising
    ``StopIteration`` when the iterator is exhausted.

        >>> p = peekable([])
        >>> p.peek('hi')
        'hi'

    peekables also offer a :meth:`prepend` method, which "inserts" items
    at the head of the iterable:

        >>> p = peekable([1, 2, 3])
        >>> p.prepend(10, 11, 12)
        >>> next(p)
        10
        >>> p.peek()
        11
        >>> list(p)
        [11, 12, 1, 2, 3]

    peekables can be indexed. Index 0 is the item that will be returned by
    :func:`next`, index 1 is the item after that, and so on:
    The values up to the given index will be cached.

        >>> p = peekable(['a', 'b', 'c', 'd'])
        >>> p[0]
        'a'
        >>> p[1]
        'b'
        >>> next(p)
        'a'

    Negative indexes are supported, but be aware that they will cache the
    remaining items in the source iterator, which may require significant
    storage.

    To check whether a peekable is exhausted, check its truth value:

        >>> p = peekable(['a', 'b'])
        >>> if p:  # peekable has items
        ...     list(p)
        ['a', 'b']
        >>> if not p:  # peekable is exhausted
        ...     list(p)
        []

    """

    def __init__(self, iterable):
        self._it = iter(iterable)
        self._cache = deque()

    def __iter__(self):
        return self

    def __bool__(self):
        try:
            self.peek()
        except StopIteration:
            return False
        return True

    def peek(self, default=_marker):
        """Return the item that will be next returned from ``next()``.

        Return ``default`` if there are no items left. If ``default`` is not
        provided, raise ``StopIteration``.

        """
        if not self._cache:
            try:
                self._cache.append(next(self._it))
            except StopIteration:
                if default is _marker:
                    raise
                return default
        return self._cache[0]

    def prepend(self, *items):
        """Stack up items to be the next ones returned from ``next()`` or
        ``self.peek()``. The items will be returned in
        first in, first out order::

            >>> p = peekable([1, 2, 3])
            >>> p.prepend(10, 11, 12)
            >>> next(p)
            10
            >>> list(p)
            [11, 12, 1, 2, 3]

        It is possible, by prepending items, to "resurrect" a peekable that
        previously raised ``StopIteration``.

            >>> p = peekable([])
            >>> next(p)
            Traceback (most recent call last):
              ...
            StopIteration
            >>> p.prepend(1)
            >>> next(p)
            1
            >>> next(p)
            Traceback (most recent call last):
              ...
            StopIteration

        """
        self._cache.extendleft(reversed(items))

    def __next__(self):
        if self._cache:
            return self._cache.popleft()

        return next(self._it)

    def _get_slice(self, index):
        # Normalize the slice's arguments
        step = 1 if (index.step is None) else index.step
        if step > 0:
            start = 0 if (index.start is None) else index.start
            stop = maxsize if (index.stop is None) else index.stop
        elif step < 0:
            start = -1 if (index.start is None) else index.start
            stop = (-maxsize - 1) if (index.stop is None) else index.stop
        else:
            raise ValueError('slice step cannot be zero')

        # If either the start or stop index is negative, we'll need to cache
        # the rest of the iterable in order to slice from the right side.
        if (start < 0) or (stop < 0):
            self._cache.extend(self._it)
        # Otherwise we'll need to find the rightmost index and cache to that
        # point.
        else:
            n = min(max(start, stop) + 1, maxsize)
            cache_len = len(self._cache)
            if n >= cache_len:
                self._cache.extend(islice(self._it, n - cache_len))

        return list(self._cache)[index]

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self._get_slice(index)

        cache_len = len(self._cache)
        if index < 0:
            self._cache.extend(self._it)
        elif index >= cache_len:
            self._cache.extend(islice(self._it, index + 1 - cache_len))

        return self._cache[index]


def _get_type(c, item):
    """
    Get the type of a column depending on its name and one example item.
    Returns None for undetermined type.
    """
    name, type = c['name'], c['type']
    if type == 'reference':
        return pl.String
    if name == 'reads':
        return pl.UInt64
    return pl.String


def deduplicator_to_table(deduplicator, barcode_hierarchy):
    """
    Transform a collections.Counter to a table.
    The counter has a tuple as keys and the readcount as values,
    the order of the columns must be the same order as the tuple-key + "reads"
    """
    barcode_hierarchy = barcode_hierarchy+ [{'name': "reads", 'type': 'count'}]
    column_names = [c['name'] for c in barcode_hierarchy]
    rows = ((*items, readcount) for items, readcount in deduplicator.items())
    rows = peekable(rows)
    try:
        first_row = rows.peek()
    except StopIteration:
        return None

    arrays = zip(*rows)
    types = [_get_type(c, value) for c, value in zip(barcode_hierarchy, first_row)]

    assert len(first_row) == len(barcode_hierarchy)

    try:
        arrays = [pl.Series(bc["name"], x, dtype=type) for x, type, bc in zip(arrays, types, barcode_hierarchy)]
    except Exception as e:
        logging.error('First row %s', first_row)
        logging.error('Column names %s', column_names)
        logging.error(f'First entry of each array: {[col[0] for col in arrays]}, types: {types}')
        raise e
    table = pl.DataFrame(arrays)
    return table

def table_to_counter(table, columns):
    """
    Transform a table to a collections.counter.
    `columns` determines the order of presence of columsn in the counter.
    """
    d = table.to_dict()
    counter = Counter(dict(zip(zip(*(d[col] for col in columns)), d["reads"])))
    return counter


def get_sectioned_iterator(inputgroup, start, end):
    if len(inputgroup.files) == 1:
        inputfile = inputgroup.files[0]
        with open_sequencing_file(inputfile.path) as f:
            if start != 0:
                f.seek(start)
            for read in f:
                yield read
                if f.tell() >= end:
                    break
    else:
        input_files = [(iter(open_sequencing_file(inputfile.path)), inputfile.prefix) for inputfile in inputgroup.files]

        running = True

        while running:
            read = {}
            for f, prefix in input_files:
                try:
                    partial_read = next(f)
                except StopIteration:
                    running = False
                    break
                for tag in partial_read:
                    read[prefix+tag] = partial_read[tag]
            if running:
                yield read


def deduplicate_section(inputgroup, section,
        regex_dictionary,
        barcode_hierarchy,
        count_qc: bool):
    logger = logging.getLogger()

    if section is None:
        start, end = 0, np.inf
    else:
        start, end = section

    regexes = regex_dictionary.get_regexes_for_group(inputgroup.group)

    feature_deduplicator : Counter = Counter()

    # aggregate some information while iterating through all reads
    processed_reads = 0                   # count processed reads
    remaining_reads = 0                   # count surviving reads
    counters_success = {bc["name"]: StringCounter() for bc in barcode_hierarchy if bc["type"] == "reference" }
    counters_fail = StringCounter()

    logging_buffer = 5000000

    logger.debug(f"Processing group {inputgroup.group}")

    opened_input_file = get_sectioned_iterator(inputgroup, start, end)

    for read in opened_input_file:
        insufficient = False
        processed_reads += 1
        if processed_reads % logging_buffer == 0:
            logger.info(f"Processed {processed_reads} reads from group {inputgroup.group}")

        features = extract_read(regexes, read)

        for bc in barcode_hierarchy:
            tag = bc["name"]
            if tag not in features or len(features[tag]) == 0:
                counters_fail[tag] += 1
                insufficient = True

        if not insufficient:
            for bc in barcode_hierarchy:
                col = bc["name"]
                if bc["min_length"] is not None:
                    if len(features[col]) < bc["min_length"]:
                        insufficient = True
                        counters_fail[col+"_length"] += 1
                if bc ["max_length"] is not None:
                    features[col] = features[col][:bc["max_length"]]

        if not insufficient:
            for bc in barcode_hierarchy:
                if bc["type"] == "reference":
                    col = bc["name"]
                    correct = deterministic_correct(features[col], bc["reference"], None, bc["threshold"])
                    features[col] = correct

                    if correct is None:
                        counters_fail.update([col+"_correction_failed"])
                        insufficient = True

        if not insufficient:
            for bc in barcode_hierarchy:
                if bc["type"] == "reference":
                    col = bc["name"]
                    counters_success[col].update([features[col]])

            data = tuple(features.get(bc["name"], "") for bc in barcode_hierarchy)
            feature_deduplicator.update([data])
            remaining_reads += 1

    logger.info(f"Finished parsing {inputgroup.group}.")

    logger.info("Sorting each sample")

    table = deduplicator_to_table(feature_deduplicator, barcode_hierarchy)
    if table is not None:
        table = table.sort([bc["name"] for bc in barcode_hierarchy])

    logger.info("demultiplexing thread finished")

    return table, (processed_reads, remaining_reads, counters_success, counters_fail) 


def extract_read(regex_dictionary, read):
    """Extract features from read

    Reads are made out of three regions of interest: The query, the b2tag and the bctag.
    Features can be scattered across these regions, e.g. part1 barcode is in the query region
    and part2 is in the b2tag.

    The function applies the user defined regex for each region and then aggregates
    the features. 
    """
    groups = {}

    for tag in read:
        re = regex_dictionary.get(tag, None)
        if re:
            match_ = re.search(read[tag])
        else:
            continue

        if not match_:
            continue

        groups.update(match_.groupdict())

    tag_stems = [tag.rstrip('0123456789') for tag in groups.keys()]

    # normalize ["sample", "sample1"] to ["sample0", "sample1"] so that sort handles this correctly
    for tag0 in tag_stems:
        if tag0 in groups:
            groups[tag0 + "0"] = groups[tag0]
            del groups[tag0]

    features = {}
    for tag in tag_stems:
        combined_tag = functools.reduce(
            lambda x, y: x + groups[y],
            sorted(name for name in groups.keys() if name.startswith(tag)),
            "",
        )
        features[tag] = combined_tag

    return features


def _combine_duplicators_polars2(main_table, partial_table, barcode_hierarchy):
    if main_table is None:
        return partial_table
    if partial_table is None:
        return main_table

    table = pl.concat([main_table, partial_table])

    table = table.group_by([bc["name"] for bc in barcode_hierarchy]).agg(pl.col("reads").sum())
    table = table.sort([bc["name"] for bc in barcode_hierarchy])
    return table


def _combine_duplicators_polars(main_table, partial_table, barcode_hierarchy):
    if main_table is None:
        return partial_table
    if partial_table is None:
        return main_table

    bchi = 0
    while bchi < len(barcode_hierarchy) and barcode_hierarchy[bchi]["type"] == "random":
        bchi += 1

    if bchi < len(barcode_hierarchy) and barcode_hierarchy[bchi]["type"] == "reference":
        anchor = pl.col(barcode_hierarchy[bchi]["name"])
        df1_anchor = collect(main_table.select(anchor.unique(maintain_order=True).alias("anchor")))
        df2_anchor = collect(partial_table.select(anchor.unique(maintain_order=True).alias("anchor")))

        anchor_values = pl.concat([df1_anchor, df2_anchor]).select(pl.col("anchor").unique(maintain_order=True))["anchor"]

        tables = []
        for a in anchor_values:
            df = (main_table.filter(anchor == a).join(partial_table.filter(anchor == a), on=[bc["name"] for bc in barcode_hierarchy if bc["name"] != "reads"], how="full", coalesce=True))
            df = df.select(*[bc["name"] for bc in barcode_hierarchy if bc["name"] != "reads"],
                                       (pl.col("reads").fill_null(0) + pl.col("reads_right").fill_null(0)).alias("reads"))
            df = df.sort([bc["name"] for bc in barcode_hierarchy])
            tables.append(df)
        return pl.concat(tables)
    else:
        table = (main_table.join(partial_table, on=[bc["name"] for bc in barcode_hierarchy if bc["name"] != "reads"], how="full", coalesce=True))
        table = table.select(*[bc["name"] for bc in barcode_hierarchy if bc["name"] != "reads"],
                                   (pl.col("reads").fill_null(0) + pl.col("reads_right").fill_null(0)).alias("reads"))

        table = table.sort([bc["name"] for bc in barcode_hierarchy])
        return table


def combine(deduplications, barcode_hierarchy):
    main_table, processed_reads, remaining_reads, counters_success, counters_fail = None, 0, 0, defaultdict(Counter), Counter()

    for i, deduplication in enumerate(deduplications):
        partial_table, (_processed_reads, _remaining_reads,  _counters_success, _counters_fail) = deduplication

        main_table = _combine_duplicators_polars(main_table, partial_table, barcode_hierarchy)

        processed_reads += _processed_reads
        remaining_reads += _remaining_reads
        counters_success = {name: counters_success[name] + Counter(_counters_success[name]) for name in set(counters_success.keys()) | set(_counters_success.keys())}
        counters_fail += _counters_fail

        del deduplication
        del partial_table

    return main_table, (processed_reads, remaining_reads, counters_success, counters_fail)


def error_correct(table, barcode_hierarchy, threshold):
    partitions = [table]
    tag_names = [tag["name"] for tag in barcode_hierarchy[1:]]
    for tag_idx, tag in enumerate(barcode_hierarchy):
        logging.debug("Error correcting barcode %s of type %s", type["name"], type["type"])
        new_partitions = []
        for partition in partitions:
            if tag["type"] == "random":
                partition = error_correct_column(partition, tag['name'], threshold)

            sort_necessary = tag['type'] == 'random'
            ps = sort_and_partition(partition, sort_columns=[tag['name']], partition_columns=[tag['name']], unwrap_if_single=False, sort=sort_necessary)
            for partition in ps.values():
                new_partitions.append(partition)
        partitions = new_partitions
    if len(partitions) > 0:
        table = pl.concat(partitions)
    return table

def error_correct_column(partition, tag, threshold):
    logging.debug("START Error correcting column %s", tag)
    tag_length = len(partition[tag][0])

    wrong_lengths = partition.filter(pl.col(tag).str.len_chars() != tag_length)

    if len(wrong_lengths) > 0:
        logging.error(f"There are {len(wrong_lengths)} many barcodes that do not have length {tag_length}, aborting")
        logging.error(f"One example: {wrong_lengths.head(1)}")
        raise ValueError(f"There are {len(wrong_lengths)} many barcodes that do not have length {tag_length}, aborting")

    tag_and_reads = partition.drop([column for column in partition.columns if not column in [tag, 'reads']])
    tag_and_reads = deduplicate_single(tag_and_reads)

    correction = calculate_corrections(tag_and_reads[tag], tag_and_reads['reads'], tag_length,  threshold)
    partition = apply_correction(partition, tag, correction)
    logging.debug("DONE Error correcting column %s", tag)
    return partition


def remove_ambigious(df, barcode_hierarchy):
    for tag in barcode_hierarchy:
        if tag['type'] == 'random':
            df = df.filter(~pl.col(tag['name']).str.contains_any(["R", "Y", "S", "W", "K", "M", "B", "D", "H", "V", "N"]))
    return df


def deduplicate_single(table):
    return table.group_by([bc for bc in table.columns if bc != "reads"]).agg(pl.col("reads").sum())


def calculate_corrections(tags, reads, tag_length, threshold):
    ec = ErrorCorrector(tag_length, 0, 0, threshold)
    tags = [(Umi(0, 0, tag), UmiCount(r, 1)) for tag, r in zip(tags, reads)]
    tags_dict = dict(tags)
    result = ec.process(tags_dict)

    corrections = result[1]
    corrected   = result[0]

    correction = {a.tag: b.tag for a, b in corrections.items()}

    return correction


def apply_correction(table, correction_column, correction):
    table = table.with_columns(pl.col(correction_column).replace_strict(correction, default=None))
    table = table.filter(~pl.col(correction_column).is_null())
    table = deduplicate_single(table)
    return table


def calculate_hopping_threshold(table, hopping_barcodes, alpha_threshold=0.05):

    columns = table.columns
    anchor_barcodes = [col for col in table.columns if col not in hopping_barcodes]
    if 'reads' in anchor_barcodes:
        anchor_barcodes.remove('reads')

    table = table.select(pl.col("reads").sum().over(anchor_barcodes).alias("n"), (pl.col("reads") >= 2).sum().over(anchor_barcodes).alias("sig"))
    table = table.with_columns(pl.when(pl.col("sig") > 0).then(1/pl.col("sig")).otherwise(pl.lit(1.0)).alias("p"))
    threshold = table.select(pl.struct(["n", "p"]).map_batches(lambda x: binom.isf(alpha_threshold, x.struct.field("n"), x.struct.field("p"))).alias("threshold"))
    return threshold


def call_hopping(table, threshold_list):

    thresholds = pl.concat([df.select(pl.col("threshold").alias(f"treshold{i}")) for i, df in enumerate(threshold_list)], how="horizontal")
    thresholds = thresholds.select(pl.max_horizontal(*[c for c in thresholds.columns]).alias("threshold"))

    df = pl.concat([table, thresholds], how="horizontal")
    df = df.filter(pl.col("reads") >= pl.col("threshold"))
    df = df.select(table.columns)

    return df


def sort_and_partition(table, sort_columns=("samplename", "reads"), partition_columns=("samplename",), unwrap_if_single=True, return_bounds=False, sort=True):
    partition_keys, partitions = _sort_and_partition(table, sort_columns, partition_columns, sort=sort)

    if not return_bounds:
        partitions = [p for p, start, end in partitions]
    d = dict(zip([tuple(key) for key in partition_keys], partitions))

    # in case of only one partition key, make keys the values instead of tuple with length 1
    if unwrap_if_single:
        if len(list(d.keys())[0]) == 1:
            d = {key[0]: value for key, value in d.items()}

    return d

def _sort_and_partition(table, sort_columns=("samplename", "reads"), partition_columns=("samplename",), sort=True):
    if isinstance(partition_columns, str):
        partition_columns = [partition_columns]

    p_col_idx = [sort_columns.index(col) for col in partition_columns]
    table_columns = table.columns

    # all partition columns must appear as sorting columns. More sorting columns are allowed.
    assert all(i != -1 for i in p_col_idx)
    # partition columns must be in same order as sort columns
    assert sorted(p_col_idx) == p_col_idx
    assert all(col in table_columns for col in sort_columns)

    table_length = len(table)

    if sort:
        table = table.sort(*sort_columns)

    partitions     = [(table, 0, len(table))]
    partition_keys = [[]]

    for col in partition_columns:
        new_partitions = []
        new_partition_keys = []

        for part, key in zip(partitions, partition_keys):
            p, orig_start, orig_length = part

            new_keys  = p[col].unique(maintain_order=True)

            start = 0
            indexes = []
            for new_key in new_keys:
                idx = p[col].slice(start).index_of(new_key)
                idx = idx + start
                indexes.append(idx)
                start = idx
            indexes.append(len(p))

            d = {}
            for new_key, (start, end) in zip(new_keys, pairwise(indexes)):
                new_partition_key = [*key, new_key]
                assert end-start >= 0
                new_partition = (p.slice(start, end-start), orig_start+start, end-start)

                new_partitions.append(new_partition)
                new_partition_keys.append(new_partition_key)

        partitions = new_partitions
        partition_keys = new_partition_keys
    assert sum(len(p[0]) for p in partitions) == table_length
    return partition_keys, partitions


def threshold_table(table, threshold):
    table = table.join(threshold, "sample")
    table = table.filter(pl.col("reads") >= pl.col("threshold"))
    return table


def _metric(a, b):
    return sum([aa != bb for aa, bb in zip(a,b)])


def deterministic_correct(barcode, references, default_value, threshold):
    # TODO do this better
    r = references.get(barcode, None)
    if r is not None:
        return r
    else:
        for reference_bc, reference_name in references.items():
            if _metric(barcode, reference_bc) <= threshold:
                return reference_name
        return default_value


def calculate_threshold(value_dict):
    th = np.inf
    for v1 in value_dict.keys():
        for v2 in value_dict.keys():
            if v1 == v2:
                continue
            th = min(th, int((_metric(v1, v2)-1)/2))
    return th



import functools

def regexes_are_valid(regexes):
    all_groups = functools.reduce(lambda x, y: x | y.groupindex.keys(), regexes, set())

    sample = [group for group in all_groups if group.startswith("sample")]
    lid = [group for group in all_groups if group.startswith("lid")]

    return len(sample) > 0 and len(lid) > 0





from bokeh.models import Plot, ColumnDataSource, CDSView
import bokeh.events
from bokeh.models import Legend, FactorRange, LabelSet, ColumnDataSource, CDSView, CustomJSFilter, CustomJS, Button
from bokeh.layouts import column, grid
from bokeh.plotting import figure
from bokeh.io import export_svg, save as save_html
import numpy as np
import polars as pl


def _make_backend_svg(p):
    if isinstance(p, Plot):
        p.output_backend = "svg"
    else:
        if hasattr(p, "children"):
            for child in p.children:
                _make_backend_svg(child)
        else:
            try:
                for child in p:
                    _make_backend_svg(child)
            except:
                pass


def save(p, output=None):
    """Save a bokeh plot as svg"""
    _make_backend_svg(p)
    if not output.endswith(".svg"):
        output += ".svg"
    export_svg(p, filename=output)


def make_pixel_filter(p, source, x_column, y_column, pixel_overlap, glyph_size):
    custom_filter = CustomJSFilter(args={'p':p, 'x_column': x_column, 'y_column':y_column, 'pixel_overlap': pixel_overlap, 'glyph_size': glyph_size}, code='''
    const frame_width = p.frame_width;
    const frame_height = p.frame_height;
    var xo = Math.inf;
    var yo = Math.inf;

    const overlap = pixel_overlap * glyph_size;
    var total_pixels = 0;

    const x_low = Math.log(p.x_range.start);
    const x_high = Math.log(p.x_range.end);
    const y_low = Math.log(p.y_range.start);
    const y_high = Math.log(p.y_range.end);

    const x_col = source.data[x_column];
    const y_col = source.data[y_column];
    const l = source.get_length();

    if ((this._cached_frame_width === frame_width) &&
        (this._cached_frame_height === frame_height) &&
        (this._cached_x_range_start === p.x_range.start) &&
        (this._cached_x_range_end === p.x_range.end) &&
        (this._cached_y_range_start === p.y_range.start) &&
        (this._cached_y_range_end === p.y_range.end) &&
        (this._cached_pixel_overlap === pixel_overlap) &&
        (this._cached_glyph_size === glyph_size)) {
        console.log("cache hit");
        return this._cached_indices;
    }

    const indices = new Array(l);

    this._cached_frame_width = frame_width;
    this._cached_frame_height = frame_height;
    this._cached_x_range_start = p.x_range.start;
    this._cached_x_range_end = p.x_range.end;
    this._cached_y_range_start = p.y_range.start;
    this._cached_y_range_end = p.y_range.end;
    this._cached_pixel_overlap = pixel_overlap;
    this._cached_glyph_size = glyph_size;
    this._cached_indices = indices;




    for (let i = 0; i < l; i++){
        var x = frame_width * (Math.log(x_col[i]) - x_low) / (x_high - x_low);
        var y = frame_height * (Math.log(y_col[i]) - y_low) / (y_high - y_low);

        if ((x < 0 ) || (x>frame_width) || (y < 0) || ( y>frame_height) ){
            indices[i] = false;
        } else {

            var dist = Math.sqrt(Math.pow(x - xo, 2) + Math.pow(y - yo, 2))

            if ((dist > overlap) || (isNaN(dist))){
                xo = x;
                yo = y;
                indices[i] = true;
                total_pixels += 1;
            } else {
                indices[i] = false;
            }
        }
    }
    console.log(total_pixels);
    return indices;
    ''')

    refresh = CustomJS(args={'filter':custom_filter, 'source':source, "p": p, "pixel_overlap": pixel_overlap, "glyph_size": glyph_size}, code="""
    setTimeout(function() {
        if (filter._may_not_draw) {
            console.log("filter may not draw");
            return;
        }

        if ((filter._cached_frame_width === p.frame_width) &&
            (filter._cached_frame_height === p.frame_height) &&
            (filter._cached_x_range_start === p.x_range.start) &&
            (filter._cached_x_range_end === p.x_range.end) &&
            (filter._cached_y_range_start === p.y_range.start) &&
            (filter._cached_y_range_end === p.y_range.end) &&
            (filter._cached_pixel_overlap === pixel_overlap) &&
            (filter._cached_glyph_size === glyph_size)) {
            console.log("refresh cache hit");
        } else {
            source.change.emit();
        }
    });""")


    p.x_range.js_on_change('start', refresh)
    p.x_range.js_on_change('end', refresh)
    p.y_range.js_on_change('start', refresh)
    p.y_range.js_on_change('end', refresh)

    disable_draw = CustomJS(args={'filter':custom_filter, 'p':p}, code="console.log('disable draw'); filter._may_not_draw = true;")
    enable_draw = CustomJS(args={'filter':custom_filter, 'p':p, 'source':source}, code="console.log('enable draw'); filter._may_not_draw = false; source.change.emit();")

    p.js_on_event(bokeh.events.PanStart, disable_draw)
    p.js_on_event(bokeh.events.PanEnd, enable_draw)

    return custom_filter





def pixel_filter(x, y, nx, ny, glyph_size, frame_width, frame_height, x_range, y_range, pixel_overlap):

    x_pixel = frame_width * (nx - x_range[0]) / (x_range[1] - x_range[0])
    y_pixel = frame_height * (ny - y_range[0]) / (y_range[1] - y_range[0])

    mask = np.ones(len(x_pixel), dtype=bool)

    xo, yo = np.inf, np.inf

    for idx, (xx, yy) in enumerate(zip(x_pixel, y_pixel)):
        dist = np.sqrt((xx-xo)**2 + (yy-yo)**2)
        if dist > pixel_overlap * glyph_size:
            xo, yo = xx, yy
        else:
            mask[idx] = False
    mask[0] = True
    mask[-1] = True
    return x[mask], y[mask]


def make_ranked(data, frame_width=200, frame_height=200, filter_pixels=False, **kwargs):
    r = data.to_numpy()
    ranks = np.arange(len(r))+1
    if filter_pixels:
        ranks, r = pixel_filter(ranks, r, np.log10(ranks), np.log10(r), 1, frame_width, frame_height, (0, np.log10(len(r))), (np.log10(np.min(r)), np.log10(np.max(r))), 1)
    source = ColumnDataSource({'Rank': ranks, 'Reads': r})
    p = figure(x_axis_type="log", y_axis_type="log", x_axis_label="Rank", y_axis_label="Reads", frame_width=frame_width, frame_height=frame_height, **kwargs)

    view = CDSView(filter=make_pixel_filter(p, source, 'Rank', 'Reads', 1.0, 1.0))
    p.scatter('Rank', 'Reads', source=source, view=view)
    return p


def make_threshold(p, max_rank, name, thresholds):
    TH = thresholds.get(name, 5)
    source_tmp = ColumnDataSource({"threshold": [TH, TH], "ranks": [1/2, max_rank*2]})
    tmp_renderer = p.line(x="ranks", y="threshold", color="red", alpha=0.5, source=source_tmp)

    source_persistent = ColumnDataSource({"name": [name, name], "threshold": [TH, TH], "ranks": [1/2, max_rank*2]})
    persistent_renderer = p.line(x="ranks", y="threshold", color="red", source=source_persistent)

    update_threshold = CustomJS(args={'source': source_tmp, 'p':p}, code="""
    source.data["threshold"][0] = cb_obj['y'];
    source.data["threshold"][1] = cb_obj['y'];
    source.change.emit();
    """)

    p.js_on_event(bokeh.events.MouseMove, update_threshold)

    hide_tmp = CustomJS(args={'renderer': tmp_renderer}, code='renderer.visible = false;')
    show_tmp = CustomJS(args={'renderer': tmp_renderer}, code='renderer.visible = true;')

    p.js_on_event(bokeh.events.MouseEnter, show_tmp)
    p.js_on_event(bokeh.events.MouseLeave, hide_tmp)

    set_threshold = CustomJS(args={'source': source_persistent, 'p':p}, code="""
    source.data["threshold"][0] = Math.round(cb_obj['y']);
    source.data["threshold"][1] = Math.round(cb_obj['y']);
    source.change.emit();
    """)

    p.js_on_event(bokeh.events.Tap, set_threshold)
    p.js_on_event(bokeh.events.DoubleTap, set_threshold)
    return source_persistent


def make_download(download_sources):
    DOWNLOAD_CODE = """
    function sources_to_csv(sources) {
        const lines = ['sample,threshold']
        for (let i = 0; i<sources.length; i++) {
            var s = sources[i].data;
            lines.push(s['name'][0] + ',' + s['threshold'][0]);
        }
        return lines.join('\\n').concat('\\n');
    }
    console.log(sources);

    const filename = download_name;
    const filetext = sources_to_csv(sources);
    const blob = new Blob([filetext], { type: 'text/csv;charset=utf-8;' });
    const link = document.createElement('a')
    link.href = URL.createObjectURL(blob)
    link.download = filename
    link.target = '_blank'
    link.style.visibility = 'hidden'
    link.dispatchEvent(new MouseEvent('click'))
    """

    button = Button(label="Download", button_type="success")
    button.js_on_click(CustomJS(args={'sources':download_sources, 'download_name': 'thresholds.csv'} ,code=DOWNLOAD_CODE))
    return button

def generate_data(df, col):
    stats = df.group_by(col).agg([pl.count().alias("count"), pl.col("reads").max()]).select([pl.col("count").max(), pl.col("reads").max()]).collect()
    max_rank = stats["count"][0]
    max_reads = stats["reads"][0]

    sorted_data = df.group_by(col).agg(pl.col("reads").sort(descending=True)).sort(pl.col(col)).collect()
    return max_rank, max_reads, sorted_data


def make_visualisation(df, col, thresholds, outpath):
    max_rank, max_reads, sorted_data = generate_data(df, col)

    ps = []
    download_sources = []
    for sample, data in zip(sorted_data["sample"], sorted_data["reads"]):
        print(sample)
        p = make_ranked(data, title=sample, x_range=(1/2, max_rank*2), y_range=(1/2, max_reads*2), filter_pixels=True)
        download_source = make_threshold(p, max_rank, sample, thresholds)
        download_sources.append(download_source)
        p.title.text_font_size = "14pt"
        p.xaxis.visible = False
        p.yaxis.visible = False
        ps.append(p)

    download_button = make_download(download_sources)

    for i, p in enumerate(ps):
        if i % 8 == 0:
            p.yaxis.visible = True
    for p in ps[-8:]:
        p.xaxis.visible = True

    save_html(column([download_button, grid(ps, ncols=8)]), outpath)





import uuid
from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
import logging
import time

from .messages import LockMsg, OKMsg, WaitMsg
from .serialization import message_to_json, parse_json_to_message, message_to_json
from .worker import Worker
from .socket_ import ReqSocket, Socket



# ==== Lock ====


class Lock:
    def __init__(self, name: str) -> None:
        self.lock = name
        self._name = uuid.uuid4().hex
        self._socket: Optional[ReqSocket] = None
        self.logger = logging.getLogger(f"remoter.lock.{name}")

    async def _get_sock_async(self) -> Socket:
        if not self._socket:

            if len(Worker.instances) == 0:
                raise Exception("Worker not instantiated")
            worker = Worker.instances[-1]
            if not worker._address:
                raise Exception("Worker context not instantiated")
            address = worker._address
            self._socket = ReqSocket()
            self.logger.debug("Connecting to: %s", address)
            self._socket.connect(address)  # type: ignore

        return self._socket  # type: ignore

    async def __aenter__(self) -> None:

        sock = await self._get_sock_async()
        msg = LockMsg("lock", self.lock, self._name)
        j = message_to_json(msg)

        while True:
            self.logger.debug("Trying to get lock: %s", j)
            await sock.async_send_json(j)
            self.logger.debug("Send json succesfully")
            j_answer = await sock.async_recv_json()
            answer = parse_json_to_message(j_answer, None)
            if isinstance(answer, OKMsg):
                self.logger.debug("Got lock")
                break
            elif isinstance(answer, WaitMsg):
                self.logger.debug("Lock not free, waiting.")
                time.sleep(answer.time)
            else:
                raise Exception("Got unexpected message on lock")

    async def __aexit__(self, type: Any, value: Any, tb: Any) -> None:
        sock = await self._get_sock_async()
        msg = LockMsg("unlock", self.lock, self._name)
        j = message_to_json(msg)

        while True:
            await sock.async_send_json(j)
            j_answer = await sock.async_recv_json()
            answer = parse_json_to_message(j_answer, None)
            if isinstance(answer, OKMsg):
                break
            elif isinstance(answer, WaitMsg):
                time.sleep(answer.time)
            else:
                raise Exception("Got unexpected message on lock")




from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
from types import TracebackType
import pickle
import base64

try:
    from typing import Literal
except:

    class MyLiteral:
        def __init__(self, t):
            self.t = t

        def __getitem__(self, *args, **kwargs):
            return self.t

    Literal = MyLiteral(str)  # type: ignore


class RemoterException(Exception):
    pass


## ==== Messages ====
#
# All Messages that can be exchanged by workers and the server


@dataclass
class FreeMsg:
    """Indicates the worker is free for more work"""

    name: str


@dataclass
class OKMsg:
    """Indicates the requested action succeeded

    Used as answer for Lock, Result and Store messages."""

    pass


@dataclass
class FailMsg:
    """Indicates the requested action failed terminally

    Used as answer for Lock, Result and Store messages."""

    name: str
    message: str


@dataclass
class QueryMsg:
    """Query properties of the server like number of workers or tasks"""

    property: str


@dataclass
class QueryAnswerMsg:
    """Answer for a query"""

    property: str


@dataclass
class TaskMsg:
    """Indicates a task to be executed by the worker"""

    name: str
    persister: str
    task_hash: str
    datetime: datetime
    function_module: str
    function_name: str
    function_hash: str
    kwargs: Dict[str, Any]
    failed: bool
    exception: Optional[tuple[BaseException, TracebackType]]
    request: bool
    time: Optional[float]

@dataclass
class FSaveMsg:
    """Saving a function"""
    datetime: datetime
    function_module: str
    function_name: str
    function_hash: str
    function_code: str

@dataclass
class ResultMsg:
    """Indicates the result of the assigned task.

    outdated, currently `Store`s communicate via StoreMsgs"""

    task_hash: str
    datetime: datetime
    function_name: str
    function_module: str
    kwargs: Dict[str, Any]
    worker_name: str
    result: Any = None


@dataclass
class WaitMsg:
    """Indicate the worker that no work is available, but more might follow

    Is used as an answer to a `FreeMsg` or to a `LockMsg`"""

    time: int


@dataclass
class QuitMsg:
    """Indicate the worker to shut down

    is used as an answer to a `FreeMsg`"""

    name: str
    client_type: str

@dataclass
class LockMsg:
    """Asks to acquire a lock"""

    operation: Literal["lock", "unlock"]
    lock: str
    name: str


@dataclass
class LoadMsg:
    """Asks to load multiple tasks"""
    task_hashs: List[str]


@dataclass
class StoreMsg:
    """Updates a Store"""

    datetime: datetime
    function_module: str
    function_name: str
    function_hash: str
    kwargs: Dict[str, Any]
    result: Union[Any, bytes]
    task_hash: str
    result_type: str
    failed: bool
    time: float

    def __repr__(self):
        kwargs = str(self.kwargs)
        if len(kwargs) > 20:
            kwargs = kwargs[:10] + "..." + kwargs[-10:]

        return f"StoreMsg(datetime={self.datetime}, function_module={self.function_module}, function_name={self.function_name}, function_hash={self.function_hash}, kwargs={kwargs}, result=..., task_hash={self.task_hash}, result_type={self.result_type}, failed={self.failed}, time={self.time})"


    def __str__(self):
        kwargs = str(self.kwargs)
        if len(kwargs) > 20:
            kwargs = kwargs[:10] + "..." + kwargs[-10:]

        return f"StoreMsg(datetime={self.datetime}, function_module={self.function_module}, function_name={self.function_name}, function_hash={self.function_hash}, kwargs={kwargs}, result=..., task_hash={self.task_hash}, result_type={self.result_type}, failed={self.failed}, time={self.time})"



@dataclass
class ExceptionMsg:
    name: str
    exception: str

@dataclass
class JoinMsg:
    name: str
    client_type: str

# Typing Stuff to make sure all messages are always handled in all code paths

Message     =          Union[QueryMsg, QueryAnswerMsg, OKMsg , FreeMsg, TaskMsg, QuitMsg, LockMsg, LoadMsg, WaitMsg, ResultMsg, StoreMsg, FailMsg, ExceptionMsg, JoinMsg, FSaveMsg]

# A serialized message indicates its type by a token, this lists the available tokens.
# Order is the same as the Message Type
MessageType =               ["QUERY", "ANSWER", "OKOK" , "FREE" ,  "TASK", "QUIT",  "LOCK",  "WAIT",  "RESULT",  "STORE",  "FAIL", "EXCEPTION", "JOIN", "FSAVE", "LOAD"]
# Type of MesssageType, used when only the tokens listed in MessageType are accectable
MessageTypeLiteral = Literal["QUERY", "ANSWER", "OKOK" , "FREE" ,  "TASK", "QUIT",  "LOCK",  "WAIT",  "RESULT",  "STORE",  "FAIL", "EXCEPTION", "JOIN", "FSAVE", "LOAD"]

# Can be used to enable exhaustive type checking in mypy, see https://hakibenita.com/python-mypy-exhaustive-checking
def assert_never(x: NoReturn) -> NoReturn:
    raise Exception("This should never happen.")



from typing import Optional
from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
from types import TracebackType
from datetime import datetime
from phantombuster.config_files import InputFile, RegexDictionary, InputGroup
from phantombuster.remoter.messages import Message, FreeMsg, OKMsg, FailMsg, QueryMsg, QueryAnswerMsg, TaskMsg, FSaveMsg, ResultMsg, WaitMsg, QuitMsg, LockMsg, LoadMsg, StoreMsg, ExceptionMsg, JoinMsg, MessageType, MessageTypeLiteral

import pickle
import json
import base64

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

def error_to_string(err: Optional[tuple[BaseException, TracebackType]]) -> Optional[str]:
    if err:
        return str(err)
        #return pickle.dumps(err, protocol=pickle.HIGHEST_PROTOCOL).decode("utf8", "surrogateescape")
    return None

def string_to_error(s: Optional[str]) -> Optional[tuple[BaseException, TracebackType]]:
    if s:
        return s, None
        return pickle.loads(s.encode("utf8", "surrogateescape"))
    return None

EXTENSIONS = {}

def register_serialisation(cls):
    EXTENSIONS[cls.__name__] = cls
    return cls

def register_serialization_extension(name, cls):
    EXTENSIONS[name] = cls

register_serialization_extension('InputFile', InputFile)
register_serialization_extension('RegexDictionary', RegexDictionary)
register_serialization_extension('InputGroup', InputGroup)

def serialize_value(value):
    if isinstance(value, Task):
        value.calc_hash()
        return ('t', value.task_hash)
    if isinstance(value, list):
        return ('l', [serialize_value(v) for v in value])
    if isinstance(value, dict):
        return ('d', {key: serialize_value(v) for key, v in value.items()})

    for name, cls in EXTENSIONS.items():
        if isinstance(value, cls):
            return ('e', {'name': name, 'value': value._to_json_e()})
    return ('a', value)

def deserialize_value(value, store):
    value_type, value = value
    if value_type == 't':
        task = store.load_from_hash(value)
        if task is None:
            raise Exception(f'Could not deserialize task with hash {value["value"]}')
        return task
    elif value_type == 'l':
        return [deserialize_value(v, store) for v in value]
    elif value_type == 'd':
        return {key: deserialize_value(v, store) for key, v in value.items()}
    elif value_type == 'e':
        cls = EXTENSIONS[value['name']]
        return cls._from_json_e(value['value'])
    elif value_type == 'a':
        return value

def serialize_kwargs(kwargs):
    if kwargs is not None:
        kwargs = {key: serialize_value(value) for key, value in kwargs.items()}
        kwargs = _clean(kwargs)
    return json.dumps(kwargs)

def deserialize_kwargs(string, store):
    kwargs = json.loads(string)
    if kwargs is not None:
        kwargs = {key: deserialize_value(value, store) for key, value in kwargs.items()}
    return kwargs

def _clean(v):
    if isinstance(v, dict):
        r = {}
        for key in sorted(v):
            r[key] = _clean(v[key])
        return r
    return v


# ==== De/Serializing Messages ====


def parse_json_to_message(j: Dict[str, Any], store: 'Store') -> Message:
    """Deserialize a json-like dict to a message"""
    if "type" not in j:
        raise Exception("Message had no declared type.")
    t: MessageTypeLiteral = j["type"]
    if t not in MessageType:
        raise Exception(f"Unknown type {t}")
    if t == "FREE":
        if "name" not in j:
            raise Exception("Free message does not contain worker name")
        return FreeMsg(j["name"])
    elif t == "OKOK":
        return OKMsg()
    elif t == "QUERY":
        return QueryMsg(j["property"])
    elif t == "ANSWER":
        return QueryAnswerMsg(j["property"])
    elif t == "FSAVE":
        return FSaveMsg(
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT) if j['datetime'] else None,
            function_module=j["function_module"],
            function_name=j["function_name"],
            function_hash=j["function_hash"],
            function_code=j["function_code"],
        )
    elif t == "TASK":
        return TaskMsg(
            name=j['name'],
            persister=j["persister"],
            task_hash=j["task_hash"],
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT) if j['datetime'] else None,
            function_module=j["function_module"],
            function_name=j["function_name"],
            function_hash=j["function_hash"],
            kwargs=deserialize_kwargs(j["kwargs"], store),
            failed=j["failed"],
            exception=string_to_error(j["exception"]),
            request=j["request"],
            time=j["time"]
        )
    elif t == "QUIT":
        return QuitMsg(name=j["name"], client_type=j["client_type"])
    elif t == "LOCK":
        return LockMsg(operation=j["op"], lock=j["lock"], name=j["name"])
    elif t == "LOAD":
        return LoadMsg(task_hashs=j["task_hashs"])
    elif t == "WAIT":
        return WaitMsg(j["time"])
    elif t == "RESULT":
        return ResultMsg(
            task_hash=j["task_hash"],
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT),
            function_module=j["function_module"],
            function_name=j["function_name"],
            kwargs=deserialize_kwargs(j["kwargs"], store),
            worker_name=j["worker_name"],
            result=j["result"],
        )
    elif t == "STORE":
        r = pickle.loads(base64.b64decode(j["result"].encode("utf-8")))
        return StoreMsg(
            datetime=datetime.strptime(j["datetime"], DATETIME_FORMAT),
            function_module=j["function_module"],
            function_name=j["function_name"],
            function_hash=j["function_hash"],
            kwargs=deserialize_kwargs(j["kwargs"], store),
            result=r,
            task_hash=j["task_hash"],
            result_type=j["result_type"],
            failed=j["failed"],
            time=j["time"],
        )
    elif t == "FAIL":
        return FailMsg(name=j["name"], message=j["message"])
    elif t == "EXCEPTION":
        return ExceptionMsg(name=j["name"], exception=j["exception"])
    elif t == "JOIN":
        return JoinMsg(name=j["name"], client_type=j["client_type"])
    else:
        assert_never(t)


def message_to_json(msg: Message) -> Dict[str, Any]:
    """Serialize a message to json-like dict"""
    if isinstance(msg, OKMsg):
        return {"type": "OKOK"}
    elif isinstance(msg, QueryMsg):
        return {"type": "QUERY", "property": msg.property}
    elif isinstance(msg, QueryAnswerMsg):
        return {"type": "ANSWER", "property": msg.property}
    elif isinstance(msg, FreeMsg):
        return {"type": "FREE", "name": msg.name}
    elif isinstance(msg, TaskMsg):
        return {
            "type": "TASK",
            "name": msg.name,
            "persister": msg.persister,
            "task_hash": msg.task_hash,
            "datetime": msg.datetime.strftime(DATETIME_FORMAT) if msg.datetime else None,
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "function_hash": msg.function_hash,
            "kwargs": serialize_kwargs(msg.kwargs),
            "failed": msg.failed,
            "exception": error_to_string(msg.exception),
            "request": msg.request,
            "time": msg.time
        }
    elif isinstance(msg, FSaveMsg):
        return {
            "type": "FSAVE",
            "datetime": msg.datetime.strftime(DATETIME_FORMAT) if msg.datetime else None,
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "function_hash": msg.function_hash,
            "function_code": msg.function_code,
        }
    elif isinstance(msg, QuitMsg):
        return {"type": "QUIT", "name": msg.name, "client_type": msg.client_type}
    elif isinstance(msg, LockMsg):
        return {"type": "LOCK", "op": msg.operation, "name": msg.name, "lock": msg.lock}
    elif isinstance(msg, LoadMsg):
        return {"type": "LOAD", "task_hashs": msg.task_hashs}
    elif isinstance(msg, WaitMsg):
        return {"type": "WAIT", "time": msg.time}
    elif isinstance(msg, ResultMsg):
        return {
            "type": "RESULT",
            "task_hash": msg.task_hash,
            "datetime": msg.datetime.strftime(DATETIME_FORMAT),
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "kwargs": serialize_kwargs(msg.kwargs),
            "worker_name": msg.worker_name,
            "result": msg.result,
        }
    elif isinstance(msg, StoreMsg):
        r = base64.b64encode(pickle.dumps(msg.result)).decode("utf-8")
        msg = {
            "datetime": msg.datetime.strftime(DATETIME_FORMAT),
            "type": "STORE",
            "function_module": msg.function_module,
            "function_name": msg.function_name,
            "function_hash": msg.function_hash,
            "kwargs": serialize_kwargs(msg.kwargs),
            "result": r,
            "task_hash": msg.task_hash,
            "result_type": msg.result_type,
            "failed": msg.failed,
            "time": msg.time,
        }
        return msg
    elif isinstance(msg, FailMsg):
        return {"type": "FAIL", "name": msg.name, "message": msg.message}
    elif isinstance(msg, ExceptionMsg):
        return {"type": "EXCEPTION", "name": msg.name, "exception": msg.exception}
    elif isinstance(msg, JoinMsg):
        return {"type": "JOIN", "name": msg.name, "client_type": msg.client_type}
    else:
        assert_never(msg)

def init_task(task):
    global Task
    Task = task



from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
import os.path
import sys
import inspect
from datetime import datetime
import logging
import traceback
import time

from .persisters import PersisterMapper, persister_factory
from .task import Task, hash_function
from .store import Store, Remote
from .async_ import async_run, async_gather
from .messages import Message, FreeMsg, QuitMsg, WaitMsg, OKMsg, TaskMsg, LockMsg, LoadMsg, ResultMsg, StoreMsg, FailMsg, ExceptionMsg, assert_never, JoinMsg, FSaveMsg, QueryMsg, QueryAnswerMsg, RemoterException
from .serialization import parse_json_to_message, message_to_json
from .socket_ import RepSocket, PullSocket
from .worker import Worker

DEFAULT_STORE_PATH = ".remoter.sqlite3"

class WorkerManager:

    def __init__(self):
        self.workers = []

        self._working_on = {}
        self._tasks_in_progress_list: List[Task] = []

    @property
    def tasks_in_progress(self):
        return self._tasks_in_progress_list

    def __len__(self):
        return len(self.workers)

    def remove(self, worker):
        if worker in self.workers:
            self.workers.remove(worker)
        if worker in self._working_on:
            del self._working_on[worker]

    def add(self, worker):
        if worker not in self.workers:
            self.workers.append(worker)

    def notice(self, worker):
        self.add(worker)

    def works_on(self, worker, task):
        self.notice(worker)
        self._working_on[worker] = task
        self._tasks_in_progress_list.append(task)

    def get_working_on(self, worker):
        return self._working_on.get(worker, None)

    def done(self, worker):
        if worker in self._working_on:
            task = self._working_on[worker]
            del self._working_on[worker]
            self._tasks_in_progress_list.remove(task)


class TaskQueue:

    def __init__(self):
        self._ready_tasks: List[Task] = []
        self._logger = logging.getLogger("remoter.server.TaskQueue")

        self._tasks_scheduled_ever = set()
        self._total_tasks = 0

        self._tasks_being_waited_on = {}
        self._task_waiting_counter = {}


    def _find_dependent_tasks(self, v):
        if isinstance(v, dict):
            l = []
            for value in v.values():
                for t in self._find_dependent_tasks(value):
                    l.append(t)
            return list(set(l))
        elif isinstance(v, list):
            l = []
            for value in v:
                for t in self._find_dependent_tasks(value):
                    l.append(t)
            return list(set(l))
        elif isinstance(v, Task):
            if v.done:
                return []
            else:
                return [v]
        else:
            return []

    def add(self, task: Task) -> None:
        if task.calc_hash() in self._tasks_scheduled_ever:
            return
        self._tasks_scheduled_ever.add(task.calc_hash())
        self._total_tasks += 1
        tasks_to_wait_for = self._find_dependent_tasks(task.kwargs)
        for twf in tasks_to_wait_for:
            if twf.calc_hash() not in self._tasks_being_waited_on:
                self._tasks_being_waited_on[twf.calc_hash()] = []
            if task not in self._tasks_being_waited_on[twf.calc_hash()]:
                self._tasks_being_waited_on[twf.calc_hash()].append(task)
        if len(tasks_to_wait_for) > 0:
            self._logger.debug("Task with hash %s needs to wait for %s many tasks", task.calc_hash(), len(tasks_to_wait_for))
            self._task_waiting_counter[task.calc_hash()] = len(tasks_to_wait_for)
        else:
            self._ready_tasks.append(task)

    def done(self, task_hash):
        if task_hash in self._tasks_being_waited_on:
            for waiting_task in self._tasks_being_waited_on[task_hash]:
                counter = self._task_waiting_counter.get(waiting_task.calc_hash(), 0) - 1
                if counter <= 0:
                    self._logger.debug("Task with hash %s has counter %s and is now being considered for tasks", waiting_task.calc_hash(), counter)
                    self._ready_tasks.append(waiting_task)
                    del self._task_waiting_counter[waiting_task.calc_hash()]
                else:
                    self._task_waiting_counter[waiting_task.calc_hash()] = counter
            del self._tasks_being_waited_on[task_hash]

    def failed(self, task_hash):
        self._ready_tasks = []

    def remove(self, task: Task) -> bool:
        if not task in self._ready_tasks:
            return False
        self._ready_tasks.remove(task)
        return True

    def get_ready_task(self):
        if len(self._ready_tasks) == 0:
            return None
        return self._ready_tasks.pop(0)

    def __len__(self):
        return len(self._ready_tasks) + len(self._task_waiting_counter)


class LoggingServer:

    def __init__(self, address):
        self._address = "tcp://127.0.0.1:12345"
        self._socket = PullSocket()
        self._running = False

    async def serve_logging_async(self) -> None:
        logger = logging.getLogger("remoter.logger")
        logger.debug("Starting logging server")
        logger.debug("Logging server listening on %s", self._address)
        self._running = True
        self._socket.listen(self._address)
        out_logger = logging.getLogger()

        try:
            logger.debug("Starting to log, status is %s", self._running)
            while self._running:
                d = await self._socket.async_recv_json()
                logger.debug("Logging something")
                record = logging.makeLogRecord(d)
                out_logger.handle(record)
        finally:
            self._socket.close()
            logger.info("Logging server shut down done")



class Server:

    instance: Optional["Server"] = None

    def __init__(
        self,
        address: str = "ipc://.remoter.ipc",
        logging_address: str = "ipc://.remoter.logging.ipc",
        storepath: str = DEFAULT_STORE_PATH,
        persister_mapper=PersisterMapper,
        enable_logging=False,
        launch_workers=None,
        workers=None,
        lazy=False,
        client_hang_around=True,
        server_hang_around=True,
        wait_time=1,
        name='server',
    ):

        if storepath is None:
            storepath = DEFAULT_STORE_PATH

        self._logger = logging.getLogger("remoter.server")
        self._schedulers: Set[str] = set()
        self._scheduler_remotes: dict[str, "Remote"] = {}
        self._address = address
        self._socket: Optional[RepSocket] = None
        self._locks: Dict[str, str] = {}
        self._running = False
        self._store = Store(storepath, name="server_store")
        self._persister_mapper = persister_mapper
        self._logging = enable_logging
        self._failed = False
        self._failing_messages: list[str] = []
        self._current_file_module = os.path.split(sys.argv[0])[-1].replace(".py", "")
        self._logging_address = logging_address
        self._launch_worker_mode = launch_workers
        self._launch_worker_number = workers
        self._lazy = lazy
        self._client_exit_on_free = not client_hang_around
        self._server_exit_on_free = not server_hang_around
        self._worker = None
        self._wait_time = wait_time
        self._should_run = False
        self._name = name
        self._outputed_status = 0

        self.task_queue = TaskQueue()
        self.worker_manager = WorkerManager()

        worker_modes = [None, "async"]
        if launch_workers not in worker_modes:
            raise ValueError(f"Launch worker must be from {worker_modes}")
        Server.instance = self

    def stop(self):
        self._logger.info("Stopping server, waiting for all clients to quit")
        self._server_exit_on_free = True
        self._client_exit_on_free = True

    def fullfill(self, task, result):
        task.result = result
        task.done = True
        task.time = -1
        task._save_force()
        self.task_queue.done(task.task_hash)

    def create_task(self, f: Callable, *args: Any, **kwargs: Any) -> Task:
        params = inspect.signature(f).parameters
        kwargs.update({param: arg for param, arg in zip(params, args)})
        module = f.__module__ if f.__module__ != "__main__" else self._current_file_module
        task = Task(None, self._store, datetime.now(), module, f.__name__, hash_function(f), kwargs)
        task.persister = self._persister_mapper(f)
        task.task_hash = task.calc_hash()
        return task

    def run(self, f: Callable, *args: Any, **kwargs: Any) -> Task:
        task = self.create_task(f, *args, **kwargs)

        if not task.is_loadable():
            self.task_queue.add(task)
        else:
            task.load()
        return task

    def start_async(self) -> None:
        if self._worker:
            async_run(self._start_server_and_worker)
        else:
            async_run(self._start_async)

    async def _start_server_and_worker(self):
        return await async_gather(self._worker._start_async, self._start_async)

    async def _start_async(self) -> None:
        self._should_run = True
        self._running = True
        tasks = [self.serve_async]
        if self._logging:
            raise NotImplementedError()
        await self.serve_async()
        self._running = False
        self._logger.info("Server shut down done")

    async def serve_async(self) -> None:
        logger = self._logger
        self._socket = sock = RepSocket()
        logger.info("Listening on %s", self._address)
        sock.listen(self._address)

        try:
            logger.debug("Serving")
            while self._should_run:
                logger.debug("Waiting for messages")
                try:
                    msg = await self.recv_async()
                except TimeoutError:
                    continue
                logger.debug("Received message at %s: %s", time.time(), msg)
                answer = self.handle_message(msg)
                logger.debug("Sending answer at %s: %s", time.time(), answer)
                await self.send_async(answer)
                self._output_status()

                if self._server_exit_on_free:
                    self._should_run = len(self.task_queue) > 0 or len(self.worker_manager) > 0 or len(self._schedulers) > 0 or len(self.worker_manager.tasks_in_progress) > 0
                    logger.info("Server should run: %s", self._should_run)
            logger.info("Server shutting down")
        finally:
            sock.close()
            self._running = False
        logger.info("Server Serving done")

    async def recv_async(self) -> Message:
        assert self._socket
        j = await self._socket.async_recv_json()
        return parse_json_to_message(j, self._store)

    async def send_async(self, msg: Message) -> None:
        j = message_to_json(msg)
        if self._socket:
            await self._socket.async_send_json(j)
        else:
            raise Exception("Can't send a message uninitialized.")

    def handle_message(self, message: Message) -> Message:
        if isinstance(message, FreeMsg):
            if self._failed:
                self.worker_manager.remove(message.name)
                self._logger.info("Worker %s is free, telling it to quit as we are in failed state.", message.name)
                return QuitMsg("server", "server")

            task = self.task_queue.get_ready_task()
            if task:
                self.worker_manager.works_on(message.name, task)
                self._logger.debug("Worker %s is free, telling it to work on the task '%s'", message.name, task.function_name)
                return task.to_task_message(self._name)

            elif self._client_exit_on_free:
                self.worker_manager.remove(message.name)
                self._logger.debug("Worker %s is free, telling it to quit as there is no more work to do", message.name)
                return QuitMsg("server", "server")
            else:
                self.worker_manager.notice(message.name)
                return WaitMsg(self._wait_time)

        elif isinstance(message, QueryMsg):
            if message.property == 'workers':
                return QueryAnswerMsg(str(len(self.worker_manager.workers)))
            else:
                return QueryAnswerMsg("None")

        elif isinstance(message, QueryAnswerMsg):
            raise Exception("Server got a QueryAnswerMsg")

        elif isinstance(message, OKMsg):
            raise Exception("Server got a OkMessage")

        elif isinstance(message, TaskMsg):
            # Scheduler schedules a task
            if self._failed:
                self._logger.info("'%s' is asking for a task, telling it to quit as we are in failed state.", message.name)
                return QuitMsg("server", "server")

            persister = persister_factory(message.persister)
            task = Task.from_task_message(message, persister, self._store)
            storemsg = task.load_metadata()
            # Task is already calculated, returning result
            if storemsg:
                self._logger.debug("'%s' asked for task, returning task with hash '%s'", message.name, task.task_hash)
                return storemsg
            # Task needs to be run, scheduler has to wait
            else:
                if message.request:
                    self._logger.debug("'%s' asked for task, scheduling '%s'", message.name, task.function_name)
                    self.task_queue.add(task)
                return WaitMsg(self._wait_time)

        elif isinstance(message, QuitMsg):
            self._logger.info("Scheduler %s quit", message.name)
            self._schedulers.remove(message.name)
            self._store._remotes.remove(self._scheduler_remotes[message.name])
            if len(self._schedulers) == 0:
                self._logger.info("No more schedulers active, stopping")
                self.stop()
            return OKMsg()

        elif isinstance(message, LockMsg):
            if message.operation == "lock":
                if message.lock not in self._locks:
                    self._locks[message.lock] = message.name
                    return OKMsg()
                else:
                    return WaitMsg(time=1)
            else:
                if message.lock in self._locks:
                    del self._locks[message.lock]
                return OKMsg()

        elif isinstance(message, LoadMsg):
            if self._failed:
                self._logger.info("'%s' is asking for a task, telling it to quit as we are in failed state.", message.name)
                return QuitMsg("server", "server")

            try:
                msg = self._store._load_metadata_one(message.task_hashs)
            except RemoterException:
                return WaitMsg(self._wait_time)
            return msg

        elif isinstance(message, ResultMsg):
            self.worker_manager.done(message.worker_name)
            self.task_queue.done(message.task_hash)
            return OKMsg()

        elif isinstance(message, WaitMsg):
            raise Exception("Server got a WaitMsg")

        elif isinstance(message, StoreMsg):
            self._store._income(message)
            return OKMsg()

        elif isinstance(message, FailMsg):
            self._failed = True
            self._failing_messages.append(message.message)
            task = self.worker_manager.get_working_on(message.name)
            self.task_queue.failed(task.task_hash)
            self.worker_manager.remove(message.name)
            self._logger.info("Client %s informed me of failure, telling it to quit", message.name)
            return QuitMsg("server", "server")

        elif isinstance(message, ExceptionMsg):
            exception = message.exception
            logging.error(exception)
            self._failing_messages.append(exception)
            self._logger.info("Client %s informed me of an exception", message.name)
            return OKMsg()

        elif isinstance(message, JoinMsg):
            self._schedulers.add(message.name)
            remote = Remote(self, message.name)
            self._scheduler_remotes[message.name] = remote
            self._store._remotes.append(remote)
            self._logger.debug("Scheduler '%s' joined", message.name)
            return OKMsg()

        elif isinstance(message, FSaveMsg):
            self._store._funcsave(message)
            return OKMsg()
        else:
            assert_never(message)

    def _launch_workers(self):
        if len(self._tasks) > 0 and self._launch_worker_mode == "async":
            self._worker = Worker(self._address, "w1")

    def _output_status(self):
        if self._outputed_status <= 0:
            status = f"Workers: {len(self.worker_manager)}, Tasks: {self.task_queue._total_tasks}, Todo: {len(self.task_queue)}"
            self._logger.info(status)
            self._outputed_status = 1000
        else:
            self._outputed_status -= 1





from dataclasses import dataclass
import polars as pl
import numpy as np



def calculate_pvalue(null, value, n):
    idx = null[n-1,:].searchsorted(value)
    return idx / null.shape[1]


def null_distribution(df, column='cells', N=1_000, max_n=950):
    a = df[column].to_numpy()
    a = np.log2(a)
    r = np.random.choice(a, (max_n, N))
    np.cumsum(r, axis=0, out=r)
    ns = np.arange(max_n) + 1
    r /= ns.reshape((max_n, 1))
    r.sort(axis=1)
    assert r.shape == (max_n, N)
    return r


def calculate_pvalues(df, column="reads", group_by=['grna'], N=1_000_000, max_n=950):
    controls = df.filter(pl.col('category') == 'control')
    to_test = df

    if len(to_test) == 0:
        raise ValueError('No test lineages')

    if len(controls) == 0:
        raise ValueError('No control lineages')


    ctrl_mean = controls[column].log(2).mean()
    to_test = to_test.group_by(*group_by).agg(pl.col(column).log(2).mean().alias('teststatistic'), pl.len().alias('lineages'))
    to_test = to_test.with_columns((pl.col('teststatistic') - ctrl_mean).alias('foldchange'))

    max_size = to_test['lineages'].max()

    if max_size < max_n:
        max_n = max_size

    null = null_distribution(controls, column=column, max_n=max_n, N=N)

    out = {'pvalue': [], 'teststatistic': [], 'foldchange': [], "lineages": []}

    for tag in group_by:
        out[tag] = []


    for row in to_test.iter_rows(named=True):
        pvalue = calculate_pvalue(null, row['teststatistic'], min(row['lineages'], max_n))
        for tag in group_by:
            out[tag].append(row[tag])
        out['teststatistic'].append(row['teststatistic'])
        out['foldchange'].append(row['foldchange'])
        out['pvalue'].append(pvalue)
        out['lineages'].append(row['lineages'])

    df = pl.DataFrame(out)
    return df





from typing import Any, Callable, Optional
import os.path
import sys
import inspect
from datetime import datetime
import logging
import uuid
import time
import threading
import functools
import time
import zmq.error

from . import globaladdress
from .persisters import PersisterMapper
from .task import Task, hash_function
from .store import Store
from .messages import Message, StoreMsg, JoinMsg, LoadMsg, WaitMsg, QuitMsg, QueryMsg, QueryAnswerMsg
from .serialization import parse_json_to_message, message_to_json
from .async_ import async_gather, async_sleep
from .socket_ import Socket, ReqSocket
from .worker import Worker
from .server import Server

class ServerManager:

    def __init__(self, address, storepath=None):
        self.address = address
        self.server = None
        self.storepath = storepath

    def start(self):
        self.thread = threading.Thread(target=self._start)
        self.thread.start()

    def _start(self):
        self.server = Server(self.address, storepath=self.storepath)
        self.server.start_async()

    def wait(self):
        self.thread.join()

    def stop(self):
        self.thread.stop()


class SingleWorker:

    def __init__(self, address, name):
        self.address = address
        self.name = name

    def start(self):
        self.thread = threading.Thread(target=self._start)
        self.thread.start()

    def _start(self):
        self.worker = Worker(self.address, self.name)
        self.worker.start_async()

    def stop(self):
        self.thread.stop()

    def wait(self):
        self.thread.join()


class WorkerManager:

    def __init__(self, address):
        self.address = address
        self.workers = []

    def start(self):
        worker = SingleWorker(self.address, name=f'worker{len(self.workers)+1}')
        self.workers.append(worker)
        worker.start()

    def wait(self):
        for worker in self.workers:
            worker.wait()

    def stop(self):
        for worker in self.workers:
            worker.stop()

class TaskCreater:

    def __init__(self, store, persister_mapper):
        self._store = store
        self._current_file_module = os.path.split(sys.argv[0])[-1].replace(".py", "")
        self._persister_mapper = persister_mapper

    @functools.cache
    def get_params(self, f):
        return inspect.signature(f).parameters

    @functools.cache
    def get_code(self, f):
        return inspect.getsource(f)

    def create_task(self, f: Callable, *args: Any, **kwargs: Any) -> Task:
        params = self.get_params(f)
        kwargs.update({param: arg for param, arg in zip(params, args)})
        module = f.__module__ if f.__module__ != "__main__" else self._current_file_module
        code = self.get_code(f)
        task = Task(None, self._store, datetime.now(), module, f.__name__, hash_function(f), kwargs, code=code)
        task.persister = self._persister_mapper(f)
        task.task_hash = task.calc_hash()
        return task



class Scheduler:

    DEFAULT_WORKERS = 0

    def __init__(self,
                 address: Optional[str] = None,
                 persister_mapper=PersisterMapper,
                 name: Optional[str] = None,
                 start_server: bool = False,
                 start_workers: int = -1,
                 save_functions: bool = False,
                 server_store_path: Optional[str] = None,
                 ):
        # If no address is given and no explicit address was ever given, use default address
        if address is None and globaladdress.global_address is None:
            address = "ipc://.remoter.ipc"
        # if explicit address was given, reuse it for later
        elif address is not None and globaladdress.global_address is None:
            globaladdress.global_address = address
        # if no explicit address is given, reuse last explicit address
        elif address is None and globaladdress.global_address is not None:
            address = globaladdress.global_address

        if start_workers == -1:
            start_workers = Scheduler.DEFAULT_WORKERS

        self.worker_manager = WorkerManager(address)
        self.server_manager = ServerManager(address, server_store_path)

        self._address = address
        if name:
            self._name = name
        else:
            self._name = uuid.uuid4().hex
        self._running = False
        self._store = Store(":memory:", name=self._name)
        self._logger = logging.getLogger(f"remoter.scheduler.{self._name}")
        self._logger.debug(f"Connecting to {address}")
        self._socket: Socket = self._create_socket()
        self._socket.connect(address)
        self._joined = False
        self._entered = 0
        self._failed = False
        self._tasks: list["Task"] = []
        self._start_server = start_server
        self._start_workers = start_workers
        self._save_functions = save_functions

        self.task_creater = TaskCreater(self._store, persister_mapper)


    def _fullfill(self, task, result):
        task.result = result
        task.done = True
        task._save_force()

    def _create_socket(self):
        return ReqSocket()

    def _join(self):
        msg = JoinMsg(self._name, "scheduler")
        msg = self._send_and_recv(msg)
        self._joined = True

    def run(self, f: Callable, *args: Any, **kwargs: Any) -> Task:
        if not self._joined:
            self._join()
        task = self.schedule(f, *args, **kwargs)
        self.gather(task)
        return task.result

    def request(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            self._join()
        task = self.task_creater.create_task(f, *args, **kwargs)
        self._logger.debug("scheduling task: %s", task.function_name)
        msg = task.to_task_message(name=self._name, request=False)
        return_msg = self._send_and_recv(msg)
        if isinstance(return_msg, StoreMsg):
            self._store._income(return_msg)
            task.calc_hash()
            task.load()
        return task

    def schedule(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            self._join()
        task = self.task_creater.create_task(f, *args, **kwargs)
        if self._save_functions:
            msg = task.to_fsave_message()
            return_msg = self._send_and_recv(msg)
        self._logger.debug("scheduling task: %s", task.function_name)
        msg = task.to_task_message(name=self._name)
        return_msg = self._send_and_recv(msg)
        if isinstance(return_msg, StoreMsg):
            self._store._income(return_msg)
            task.calc_hash()
        else:
            self._tasks.append(task)
        return task

    def wait_all(self, tasks, unpack=False):
        return [self.wait(task, unpack=unpack) for task in tasks]

    def wait(self, task, unpack=False, return_after_wait=False):
        return self._wait_for(task, unpack=unpack, return_after_wait=return_after_wait)

    def _wait_for(self, task, unpack=False, return_after_wait=False):
        self._logger.debug("Waiting for completion of task: %s", task)
        task.calc_hash()
        while not task.is_loadable():
            msg = task.to_task_message(self._name)
            msg = self._send_and_recv(msg)

            if isinstance(msg, StoreMsg):
                self._store._income(msg)
            elif isinstance(msg, WaitMsg):
                time.sleep(msg.time)
                if return_after_wait:
                    return msg.time
            elif isinstance(msg, QuitMsg):
                self._failed = True
                self._joined = False
                return None
            else:
                raise Exception(f"Scheduler received unexpected message: {msg}")
        self._logger.debug("Successfully waited for completion of task: %s", task)
        if unpack:
            task.calc_hash()
            task.load()
            return task.result
        else:
            return task

    def _wait_for_many(self, tasks, unpack=False, return_after_wait=False):
        for task in tasks:
            task.calc_hash()
        remaining_tasks = tasks

        while len(remaining_tasks) > 0:
            msg = LoadMsg([task.task_hash for task in remaining_tasks])
            msg = self._send_and_recv(msg)

            if isinstance(msg, WaitMsg):
                self._logger.info(f"Tasks {len(remaining_tasks)} / {len(tasks)} done")
                time.sleep(msg.time)
                if return_after_wait:
                    return msg.time
            elif isinstance(msg, StoreMsg):
                self._store._income(msg)
                remaining_tasks = [task for task in remaining_tasks if not task.task_hash == msg.task_hash]
            elif isinstance(msg, QuitMsg):
                self._failed = True
                self._joined = False
                return None
            else:
                raise Exception(f"Scheduler received unexpected message: {msg}")

        if unpack:
            for task in tasks:
                task.calc_hash()
                task.load()
            return [task.result for task in tasks]
        else:
            return tasks

    def query_workers(self):
        self._logger.debug('Quering for number of workers')
        msg = QueryMsg("workers")

        answered = False

        workers = None

        while not answered:
            msg = self._send_and_recv(msg)

            if isinstance(msg, WaitMsg):
                time.sleep(msg.time)
            elif isinstance(msg, QuitMsg):
                self._failed = True
                self._joined = False
                return None
            elif isinstance(msg, QueryAnswerMsg):
                answered = True
                if msg.property != "None":
                    workers = int(msg.property)

        return workers

    def gather(self, *tasks):
        if self._joined:
            self._wait_for_many(tasks, unpack=False)
            rs = [task.load() for task in tasks]
            for task in tasks:
                if task.failed:
                    err, tb = task.exception
                    raise err.with_traceback(tb)
            return rs
        else:
            if all(task.is_loadable() for task in tasks):
                rs = [task.load() for task in tasks]
                for task in tasks:
                    if task.failed:
                        err, tb = task.exception
                        raise err.with_traceback(tb)
                return rs
            else:
                raise Exception("Not all tasks are loadable and I'm not connected to a server")

    def _send_and_recv(self, msg: Message) -> Message:
        try:
            self._send(msg)
        except KeyboardInterrupt as e:
            try:
                self._send(msg)
            except zmq.error.ZMQError:
                pass
            try:
                answer = self._recv()
            except zmq.error.ZMQError:
                pass
            raise e
        try:
            answer = self._recv()
        except KeyboardInterrupt as e:
            raise e
        return answer 

    def _recv(self) -> Message:
        if self._socket is not None:
            self._logger.debug("Waiting for message sync")
            msg_json = self._socket.recv_json()
            msg = parse_json_to_message(msg_json, self._store)
            self._logger.debug(f"Sync Received message {time.time()}: {msg}")
            return msg
        else:
            raise Exception("Scheduler has to connect before being able to receive messages")

    def _send(self, msg: Message, force=False) -> None:
        if self._socket:
            j = message_to_json(msg)
            self._logger.debug(f"Sync Sending message at {time.time()}: {msg}")
            if force:
                try:
                    self._socket.send_json(j)
                except zmq.error.ZMQError:
                    self._socket.recv()
                    self._socket.send_json(j)
            else:
                self._socket.send_json(j)
        else:
            raise Exception("Scheduler has to connect before being able to send messages")

    def __enter__(self):
        if self._entered == 0:
            if self._start_server:
                self.server_manager.start()
            for i in range(self._start_workers):
                self.worker_manager.start()
            self._join()
        self._entered += 1
        return self

    def __exit__(self, exc_type, exc, tb):
        if self._entered == 1:
            self._logger.debug("%s, %s, %s", exc_type, exc, tb)
            self._logger.debug("Scheduler exited, sending exit msg to server sync")
            self.exit(force=True)
            self._logger.debug("Notified server of exit, exiting now")
            self.server_manager.wait()
            self.worker_manager.wait()
        self._entered -= 1
        if exc:
            raise exc
        return self

    def exit(self, force=False):
        if not force:
            for task in self._tasks:
                self._wait_for(task)
        self._send(QuitMsg(self._name, "scheduler"), force=True)
        self._recv()
        self._joined = False
        self._running = False

    async def _async_join(self):
        msg = JoinMsg(self._name, "scheduler")
        await self._async_send(msg)
        msg = await self._async_recv()
        self._joined = True

    async def async_run(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            await self._async_join()
        task = await self.async_schedule(f, *args, **kwargs)
        await task.wait()
        return task.result

    async def async_schedule(self, f: Callable, *args: Any, **kwargs: Any) -> Any:
        if not self._joined:
            await self._async_join()
        task = self.task_creater.create_task(f, *args, **kwargs)
        if self._save_functions:
            msg = task.to_fsave_message()
            await self._async_send(msg)
            return_msg = await self._async_recv()
        self._logger.debug("scheduling task: %s", task.function_name)
        msg = task.to_task_message(self._name)
        await self._async_send(msg)
        return_msg = await self._async_recv()
        if isinstance(return_msg, StoreMsg):
            self._store._income(return_msg)
            task.calc_hash()
            task.load()
        self._tasks.append(task)
        return task

    async def _async_wait_for(self, task):
        self._logger.debug("Waiting for completion of task: %s", task)
        task.calc_hash()

        while not task.is_loadable():
            msg = task.to_task_message(self._name)
            j = message_to_json(msg)
            self._logger.debug(f"Sending message: {msg}")
            await self._socket.async_send_json(j)

            self._logger.debug("Waiting for message async")
            msg_json = await self._socket.async_recv_json()
            msg = parse_json_to_message(msg_json, self._store)
            self._logger.debug(f"Received message: {msg}")

            if isinstance(msg, StoreMsg):
                self._store._income(msg)
                task.calc_hash()
                task.load()
            elif isinstance(msg, WaitMsg):
                await async_sleep(msg.time)
            elif isinstance(msg, QuitMsg):
                self._failed = True
                self._joined = False
                return None
            else:
                raise Exception(f"Scheduler received unexpected message: {msg}")
        self._logger.debug("Successfully waited for completion of task: %s", task)
        return task.result

    async def async_gather(self, *tasks):
        rs = []
        for task in tasks:
            r = await self._async_wait_for(task)
            rs.append(r)
        for task in tasks:
            if task.failed:
                raise Exception(task.result)
        return rs

    async def _async_recv(self) -> Message:
        if self._socket is not None:
            self._logger.debug("Waiting for message async")
            msg_json = await self._socket.async_recv_json()
            msg = parse_json_to_message(msg_json, self._store)
            self._logger.debug(f"Received message: {msg}")
            return msg
        else:
            raise Exception("Scheduler has to connect before being able to receive messages")

    async def _async_send(self, msg: Message) -> None:
        if self._socket:
            j = message_to_json(msg)
            self._logger.debug(f"Sending message: {msg}")
            await self._socket.async_send_json(j)
        else:
            raise Exception("Scheduler has to connect before being able to send messages")

    async def __aenter__(self):
        if self._entered == 0:
            if self._start_server:
                start_server(self._address)
            for i in range(self._start_workers):
                start_worker(self._address, i)
            self._logger.debug("Scheduler entered async, sending join msg to server async")
            await self._async_join()
            self._logger.debug("Joined async")
        self._entered += 1
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._entered == 1:
            self._logger.error("%s, %s, %s", exc_type, exc, tb)
            self._logger.debug("Scheduler exited async, sending exit msg to server async")
            await self.async_exit()
            self._logger.debug("Notified server of exit, exiting now")
        self._entered -= 1
        if exc:
            raise exc
        return self

    async def async_exit(self):
        rs = []
        for task in self._tasks:
            r = await self._async_wait_for(task)
            rs.append(r)
        await self._async_send(QuitMsg(self._name, "scheduler"))
        await self._async_recv()
        self._joined = False
        self._running = False
        return rs
